import inspect
import struct
import time
from datetime import datetime
from datetime import datetime as dt, timedelta as dtd

import bson
import numpy as np
import pandas as pd
import pymongo
import pytest
from mock import Mock, patch
from pandas.testing import assert_frame_equal, assert_series_equal
from pymongo.errors import OperationFailure
from pymongo.server_type import SERVER_TYPE

import arctic
from arctic import VERSION_STORE, PandasDataFrameStore, PandasSeriesStore
from arctic._config import FwPointersCfg, FW_POINTERS_REFS_KEY
from arctic._util import mongo_count, get_fwptr_config
from arctic.date import DateRange
from arctic.date._mktz import mktz
from arctic.exceptions import NoDataFoundException, DuplicateSnapshotException, ArcticException
from arctic.store import _version_store_utils
from arctic.store import version_store
from tests.unit.serialization.serialization_test_data import _mixed_test_data
from ...util import read_str_as_pandas
from ..test_utils import enable_profiling_for_library
from tests.util import assert_frame_equal_, assert_series_equal_

ts1 = read_str_as_pandas("""         times | near
                   2012-09-08 17:06:11.040 |  1.0
                   2012-10-08 17:06:11.040 |  2.0
                   2012-10-09 17:06:11.040 |  2.5
                   2012-11-08 17:06:11.040 |  3.0""")

ts2 = read_str_as_pandas("""         times | near
                   2012-09-08 17:06:11.040 |  1.0
                   2012-10-08 17:06:11.040 |  4.0
                   2012-10-09 17:06:11.040 |  4.5
                   2012-10-10 17:06:11.040 |  5.0
                   2012-11-08 17:06:11.040 |  3.0""")

ts1_append = read_str_as_pandas("""  times | near
                   2012-09-08 17:06:11.040 |  1.0
                   2012-10-08 17:06:11.040 |  2.0
                   2012-10-09 17:06:11.040 |  2.5
                   2012-11-08 17:06:11.040 |  3.0
                   2012-11-09 17:06:11.040 |  3.0""")


symbol = 'TS1'


class FwPointersCtx:
    def __init__(self, value_to_test, do_reconcile=True):
        self.value_to_test = value_to_test
        self.do_reconcile = do_reconcile

    def __enter__(self):
        self.orig_value = arctic.store._ndarray_store.ARCTIC_FORWARD_POINTERS_CFG
        arctic.store._ndarray_store.ARCTIC_FORWARD_POINTERS_CFG = self.value_to_test

        self.reconcile_orig_value = arctic.store._ndarray_store.ARCTIC_FORWARD_POINTERS_RECONCILE
        arctic.store._ndarray_store.ARCTIC_FORWARD_POINTERS_RECONCILE = self.do_reconcile

    def __exit__(self, *args):
        arctic.store._ndarray_store.ARCTIC_FORWARD_POINTERS_CFG = self.orig_value
        arctic.store._ndarray_store.ARCTIC_FORWARD_POINTERS_RECONCILE = self.reconcile_orig_value


from pymongo.cursor import _QUERY_OPTIONS
from pymongo.message import query as __query
def _query(allow_secondary, library_name):
    def _internal_query(options, *args, **kwargs):
        coll_name = args[0]
        data_coll_name = 'arctic_{}'.format(library_name)
        versions_coll_name = data_coll_name + '.versions'
        if allow_secondary and coll_name in (data_coll_name, versions_coll_name):
            # Reads to the Version and Chunks collections are allowed to slaves
            assert bool(options & _QUERY_OPTIONS['slave_okay']) == allow_secondary, "{}: options:{}".format(coll_name, options)
        elif '.$cmd' not in coll_name:
            # All other collections we force PRIMARY read.
            assert bool(options & _QUERY_OPTIONS['slave_okay']) == False, "{}: options:{}".format(coll_name, options)
        return __query(options, *args, **kwargs)
    return _internal_query


# MongoDB always sets slaveOk when talking to a single server.
# Pretend we're a mongos for the tests that care...
#
# A _Query's slaveOk bit is already set for queries with non-primary
# read preference. If this is a direct connection to a mongod, override
# and *always* set the slaveOk bit. See bullet point 2 in
# server-selection.rst#topology-type-single.
# set_slave_ok = (
#     topology.description.topology_type == TOPOLOGY_TYPE.Single
#    and server.description.server_type != SERVER_TYPE.Mongos)


def test_store_item_new_version(library, library_name):
    with patch('pymongo.message.query', side_effect=_query(False, library_name)), \
         patch('pymongo.server_description.ServerDescription.server_type', SERVER_TYPE.Mongos):
        library.write(symbol, ts1)
        coll = library._collection
        count = mongo_count(coll)
        assert mongo_count(coll.versions) == 1

        # No change to the TS
        library.write(symbol, ts1, prune_previous_version=False)
        assert mongo_count(coll) == count
        assert mongo_count(coll.versions) == 2


@pytest.mark.xfail(reason="mongo/pymongo codepaths have changed, query no longer called for this")
def test_store_item_read_preference(library_secondary, library_name):
    with patch('arctic.arctic.ArcticLibraryBinding.check_quota'), \
         patch('pymongo.message.query', side_effect=_query(False, library_name)) as query, \
         patch('pymongo.server_description.ServerDescription.server_type', SERVER_TYPE.Mongos):
        # write an item
        library_secondary.write(symbol, ts1)
        library_secondary.write(symbol, ts1_append, prune_previous_version=False)
        # delete an individual version
        library_secondary._delete_version(symbol, 1)
    # delete the item entirely
    library_secondary.delete(symbol)
    assert query.call_count > 0


@pytest.mark.xfail(reason="mongo/pymongo codepaths have changed, query no longer called for this")
def test_read_item_read_preference_SECONDARY(library_secondary, library_name):
    # write an item
    library_secondary.write(symbol, ts1)
    with patch('pymongo.message.query', side_effect=_query(True, library_name)) as query, \
         patch('pymongo.server_description.ServerDescription.server_type', SERVER_TYPE.Mongos):
        library_secondary.read(symbol)
    assert query.call_count > 0


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def _test_query_falls_back_to_primary(library_secondary, library_name, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        allow_secondary = [True]
        def _query(options, *args, **kwargs):
            # If we're allowing secondary read and an error occurs when reading a chunk.
            # We should attempt a call to primary only subsequently.
            # In newer MongoDBs we query <database>.$cmd rather than <database>.<collection>
            if args[0].startswith('arctic_{}.'.format(library_name.split('.')[0])) and \
                        bool(options & _QUERY_OPTIONS['slave_okay']) == True:
                allow_secondary[0] = False
                raise OperationFailure("some_error")
            return __query(options, *args, **kwargs)

        library_secondary.write(symbol, ts1)
        with patch('pymongo.message.query', side_effect=_query), \
             patch('pymongo.server_description.ServerDescription.server_type', SERVER_TYPE.Mongos):
            assert library_secondary.read(symbol) is not None
        # We raised at least once on a secondary read
        assert allow_secondary[0] == False


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_store_item_metadata(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1, metadata={'key': 'value'})

        after = library.read(symbol)

        assert after.metadata['key'] == 'value'
        assert after.version
        assert_frame_equal_(after.data, ts1)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_read_metadata(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1, metadata={'key': 'value'})

        after = library.read_metadata(symbol)

        assert after.metadata['key'] == 'value'
        assert after.version
        assert after.data is None


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_read_metadata_newer_version_with_lower_id(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        now_timestamp = int(time.time())
        now = struct.pack(">i", now_timestamp)
        old_id = bson.ObjectId(now + b"\x00\x00\x00\x00\x00\x00\x00\x00")
        new_id = bson.ObjectId(now + b"\x00\x00\x00\x00\x00\x00\x00\x01")
        object_id_class = Mock()
        object_id_class.from_datetime = bson.ObjectId.from_datetime

        object_id_class.return_value = new_id
        with patch("bson.ObjectId", object_id_class):
            library.write(symbol, ts1)

        library.snapshot('s1')

        object_id_class.return_value = old_id
        with patch("bson.ObjectId", object_id_class):
            library.write(symbol, ts2)

        now_dt = datetime.fromtimestamp(now_timestamp)
        assert library.read_metadata(symbol, as_of=now_dt).version == 2


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_read_metadata_throws_on_deleted_symbol(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1, metadata={'key': 'value'})
        library.delete(symbol)

        with pytest.raises(NoDataFoundException):
            library.read_metadata(symbol)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_store_item_and_update(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        coll = library._collection

        # Store the first timeseries
        none = datetime.now()
        time.sleep(1)
        library.write(symbol, ts1)
        original = datetime.now()

        # Assertions:
        assert mongo_count(coll.versions) == 1
        assert_frame_equal_(library.read(symbol).data, ts1)

        # Update the TimeSeries
        time.sleep(1)
        library.write(symbol, ts2, prune_previous_version=False)
        recent = datetime.now()

        assert mongo_count(coll.versions) == 2
        assert_frame_equal_(library.read(symbol).data, ts2)

        # Get the different versions of the DB
        with pytest.raises(NoDataFoundException):
            library.read(symbol, as_of=none)
        assert_frame_equal_(library.read(symbol, as_of=original).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=recent).data, ts2)

        # Now push back in the original version
        time.sleep(1)
        library.write(symbol, ts1, prune_previous_version=False)

        assert mongo_count(coll.versions) == 3
        assert_frame_equal_(library.read(symbol).data, ts1)

        # Get the different versions of the DB
        with pytest.raises(NoDataFoundException):
            library.read(symbol, as_of=none)
        assert_frame_equal_(library.read(symbol, as_of=original).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=recent).data, ts2)
        assert_frame_equal_(library.read(symbol, as_of=datetime.now()).data, ts1)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_append_update(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1)
        library.snapshot('snap')

        coll = library._collection

        # Assertions:
        assert mongo_count(coll.versions) == 1
        assert_frame_equal_(library.read(symbol).data, ts1)

        # Append an item
        dts = list(ts1.index)
        dts.append(dts[-1] + dtd(days=1))
        values = list(ts1.near.values)
        values.append(47.)
        ts2 = pd.DataFrame(index=dts, data=values, columns=ts1.columns)
        ts2.index.name = ts1.index.name

        # Saving ts2 shouldn't create any new chunks.  Instead it should
        # reuse the last chunk.
        library.write(symbol, ts2, prune_previous_version=False)
        assert mongo_count(coll.versions) == 2
        assert_frame_equal_(library.read(symbol, as_of='snap').data, ts1)
        assert_frame_equal_(library.read(symbol).data, ts2)

        # We should be able to save a smaller timeseries too
        # This isn't likely to happen, so we don't care too much about space saving
        # just make sure we get it right.
        library.write(symbol, ts1, prune_previous_version=False)
        assert_frame_equal_(library.read(symbol, as_of=1).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=2).data, ts2)
        assert_frame_equal_(library.read(symbol, as_of=3).data, ts1)

        # Append an item, and add a whole new chunk
        dts = list(ts2.index)
        dts.append(dts[-1] + dtd(days=1))
        dts.append(dts[-1] + dtd(days=40))
        values = list(ts2.near.values)
        values.append(47.)
        values.append(53.)
        ts3 = pd.DataFrame(index=dts, data=values, columns=ts1.columns)
        ts3.index.name = ts1.index.name

        library.write(symbol, ts3, prune_previous_version=False)
        assert_frame_equal_(library.read(symbol, as_of=1).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=2).data, ts2)
        assert_frame_equal_(library.read(symbol, as_of=3).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=4).data, ts3)

        library.write(symbol, ts3, prune_previous_version=False)
        assert_frame_equal_(library.read(symbol, as_of=1).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=2).data, ts2)
        assert_frame_equal_(library.read(symbol, as_of=3).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=4).data, ts3)
        assert_frame_equal_(library.read(symbol, as_of=5).data, ts3)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_append(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.append(symbol, ts1, upsert=True)
        library.append(symbol, ts1_append, upsert=True)
        assert len(library.read(symbol).data) == len(ts1) + len(ts1_append)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_append_should_overwrite_after_delete(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.append(symbol, ts1, upsert=True)
        library.append(symbol, ts1_append, upsert=True)
        assert len(library.read(symbol).data) == len(ts1) + len(ts1_append)
        library.delete(symbol)
        library.append(symbol, ts1_append, upsert=True)
        assert len(library.read(symbol).data) == len(ts1_append)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_append_empty_ts(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        data = library.append(symbol, ts1, upsert=True)
        assert(data.version == 1)
        data = library.append(symbol, pd.DataFrame(), upsert=True)
        assert(data.version == 1)
        assert len(library.read(symbol).data) == len(ts1)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_append_corrupted_new_version(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        to_append = read_str_as_pandas("""  times | near
                          2012-11-09 17:06:11.040 |  30.0""")
        to_append_2 = read_str_as_pandas("""  times | near
                          2012-11-10 17:06:11.040 |  40.0""")
        library.write(symbol, ts1)
        # Append version
        library.append(symbol, to_append)
        # The append went wrong, and the new version document (written last), not available
        library._versions.find_one_and_delete({'symbol': symbol}, sort=[('version', pymongo.DESCENDING)])

        # Should still be able to append new data
        library.append(symbol, to_append_2, upsert=True)
        assert library.read(symbol).data['near'][-1] == 40.
        assert len(library.read(symbol).data) == len(ts1) + 1


def test_query_version_as_of_int(library):
    # Store the first timeseries
    library.write(symbol, ts1)
    library.write(symbol, ts2, prune_previous_version=False)

    assert_frame_equal_(library.read(symbol, as_of=1).data, ts1)
    assert_frame_equal_(library.read(symbol).data, ts2)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_list_version(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        assert len(list(library.list_versions(symbol))) == 0
        dates = [None, None, None]
        now = dt.utcnow().replace(tzinfo=mktz('UTC'))
        for x in range(len(dates)):
            dates[x] = now - dtd(minutes=130 - x)
            with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(dates[x])):
                library.write(symbol, ts1, prune_previous_version=False)
        assert len(list(library.list_versions(symbol))) == 3

        library.write(symbol, ts1, prune_previous_version=True)
        assert len(list(library.list_versions(symbol))) >= 2

        versions = list(library.list_versions(symbol))
        for i, x in enumerate([4, 3]):
            assert versions[i]['symbol'] == symbol
            assert versions[i]['date'] >= dates[i]
            assert versions[i]['version'] == x


def test_list_version_deleted(library):
    assert len(library.list_versions(symbol)) == 0
    library.write(symbol, ts1, prune_previous_version=False)
    assert len(library.list_versions(symbol)) == 1
    # Snapshot the library so we keep the sentinel version
    library.snapshot('xxx', versions={symbol: 1})
    library.delete(symbol)
    versions = library.list_versions(symbol)
    assert len(versions) == 2
    assert versions[0]['symbol'] == symbol
    assert versions[0]['version'] == 2
    assert versions[0]['snapshots'] == []
    assert versions[0]['deleted'] == True

    assert versions[1]['symbol'] == symbol
    assert versions[1]['version'] == 1
    assert versions[1]['deleted'] == False
    assert versions[1]['snapshots'] == ['xxx']


def test_list_version_latest_only(library):
    assert len(list(library.list_versions(symbol))) == 0
    dates = [None, None, None]
    now = dt.utcnow().replace(tzinfo=mktz('UTC'))
    for x in range(len(dates)):
        dates[x] = now - dtd(minutes=20 - x)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(dates[x])):
            library.write(symbol, ts1, prune_previous_version=False)
    assert len(list(library.list_versions(symbol))) == 3

    library.write(symbol, ts1, prune_previous_version=True)
    assert len(list(library.list_versions(symbol, latest_only=True))) == 1

    versions = list(library.list_versions(symbol))
    for i, x in enumerate([4, ]):
        assert versions[i]['symbol'] == symbol
        assert versions[i]['date'] >= dates[i]
        assert versions[i]['version'] == x


def test_list_version_snapshot(library):
    library.write('A', ts1)
    library.snapshot('one')
    library.write('B', ts2)
    library.snapshot('two')
    library.write('A', ts2)
    library.snapshot('three')
    library.write('C', ts2)

    assert set(x['symbol'] for x in library.list_versions()) \
        == set(['A', 'B', 'C'])

    assert set(x['symbol'] for x in library.list_versions(snapshot='one')) \
        == set(['A'])

    assert set(x['symbol'] for x in library.list_versions(snapshot='two')) \
        == set(['A', 'B'])

    assert set(x['symbol'] for x in library.list_versions(snapshot='three')) \
        == set(['A', 'B'])

    assert [x['snapshots'] for x in library.list_versions(symbol='A')] \
        == [['three', ], ['one', 'two']]

    assert [x['snapshots'] for x in library.list_versions(symbol='B')] \
        == [['two', 'three']]

    assert all('parent' not in x for x in library.list_versions(symbol='C'))


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_delete_versions(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1)
        library.write(symbol, ts2, prune_previous_version=False)
        library.write(symbol, ts1, prune_previous_version=False)
        library.write(symbol, ts2, prune_previous_version=False)

        coll = library._collection

        # Delete version 1 (ts1)
        library._delete_version(symbol, 1)
        assert_frame_equal_(library.read(symbol, as_of=2).data, ts2)
        assert_frame_equal_(library.read(symbol, as_of=3).data, ts1)

        library._delete_version(symbol, 2)
        assert_frame_equal_(library.read(symbol, as_of=3).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=4).data, ts2)

        library._delete_version(symbol, 3)
        assert_frame_equal_(library.read(symbol).data, ts2)

        library._delete_version(symbol, 4)
        assert mongo_count(coll) == 0


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_delete_bson_versions(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        coll = library._collection

        a = [{'a': 'b'}]
        c = [{'c': 'd'}]
        library.write(symbol, a)
        library.write(symbol, c, prune_previous_version=False)
        library.write(symbol, a, prune_previous_version=False)
        library.write(symbol, c, prune_previous_version=False)
        assert mongo_count(coll.versions) == 4

        library._delete_version(symbol, 1)
        assert library.read(symbol, as_of=2).data == c
        assert library.read(symbol, as_of=3).data == a
        assert mongo_count(coll.versions) == 3

        library._delete_version(symbol, 2)
        assert library.read(symbol, as_of=3).data == a
        assert library.read(symbol, as_of=4).data == c
        assert mongo_count(coll.versions) == 2

        library._delete_version(symbol, 3)
        assert mongo_count(coll.versions) == 1
        assert library.read(symbol).data == c

        library._delete_version(symbol, 4)
        assert mongo_count(coll.versions) == 0


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_read_none_does_not_exception(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, None)
        assert library.read(symbol).data is None


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_delete_item_has_symbol(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1)
        library.write(symbol, ts2, prune_previous_version=False)
        library.write(symbol, ts1, prune_previous_version=False)
        library.write(symbol, ts2, prune_previous_version=False)

        library.delete(symbol)
        for version in (1, 2, 3, 4, None):
            with pytest.raises(NoDataFoundException):
                library.read(symbol, version)

        # Has symbol returns false - this should really be has_data
        assert not library.has_symbol(symbol)
        assert symbol not in library.list_symbols()
        assert [x['version'] for x in library.list_versions(symbol)] == []


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_delete_item_snapshot(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1)
        library.write(symbol, ts2, prune_previous_version=False)
        library.write(symbol, ts1, prune_previous_version=False)
        library.snapshot('snap')
        library.write(symbol, ts2, prune_previous_version=False)

        library.delete(symbol)

        for version in (1, 2, 4, None):
            with pytest.raises(NoDataFoundException):
                library.read(symbol, version)

        # Can get the version out of the snapshots
        assert_frame_equal_(library.read(symbol, 'snap').data, ts1)
        assert_frame_equal_(library.read(symbol, 3).data, ts1)

        assert not library.has_symbol(symbol)
        assert not library.has_symbol(symbol, as_of=2)
        assert library.has_symbol(symbol, as_of=3)
        assert symbol in library.list_symbols(all_symbols=True)
        assert symbol in library.list_symbols(snapshot='snap')
        assert symbol not in library.list_symbols()
        assert sorted([x['version'] for x in library.list_versions(symbol)]) == [3, 5]

        # Should be able to create another snapshot
        library.snapshot('snap2')
        with pytest.raises(NoDataFoundException):
            library.read(symbol, 'snap2')
        assert_frame_equal_(library.read(symbol, 'snap').data, ts1)
        assert symbol in library.list_symbols(snapshot='snap')
        assert symbol not in library.list_symbols(snapshot='snap2')


def test_has_symbol(library):
    assert not library.has_symbol(symbol)
    library.write(symbol, ts1)
    assert library.has_symbol(symbol)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_snapshot(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1)
        library.snapshot('current')
        library.write(symbol, ts2)
        assert_frame_equal_(library.read(symbol, as_of='current').data, ts1)
        assert_frame_equal_(library.read(symbol).data, ts2)
        versions = library.list_versions(symbol)
        assert versions[0]['snapshots'] == []
        assert versions[1]['snapshots'] == ['current']

        library.snapshot('new')
        assert_frame_equal_(library.read(symbol, as_of='current').data, ts1)
        assert_frame_equal_(library.read(symbol, as_of='new').data, ts2)
        assert_frame_equal_(library.read(symbol).data, ts2)
        versions = library.list_versions(symbol)
        assert versions[0]['snapshots'] == ['new']
        assert versions[1]['snapshots'] == ['current']

        # Replace the current version, and the snapshot shouldn't be deleted
        library.write(symbol, ts1, prune_previous_version=True)
        assert_frame_equal_(library.read(symbol, as_of='current').data, ts1)
        assert_frame_equal_(library.read(symbol, as_of='new').data, ts2)
        assert_frame_equal_(library.read(symbol).data, ts1)
        versions = library.list_versions(symbol)
        assert versions[0]['snapshots'] == []
        assert versions[1]['snapshots'] == ['new']
        assert versions[2]['snapshots'] == ['current']


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_snapshot_with_versions(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        """ Test snapshot of write versions consistency """
        library.write(symbol, ts1)
        library.write(symbol, ts2)

        # ensure snapshot of previous version is taken
        library.snapshot('previous', versions={symbol: 1})
        versions = library.list_versions(symbol)
        assert versions[0]['snapshots'] == []
        assert versions[1]['snapshots'] == ['previous']
        assert_frame_equal_(library.read(symbol, as_of='previous').data, ts1)

        # ensure new snapshots are ordered after previous ones
        library.snapshot('new')
        versions = library.list_versions(symbol)
        assert versions[0]['snapshots'] == ['new']
        assert versions[0]['version'] == 2
        assert_frame_equal_(library.read(symbol, as_of='new').data, ts2)

        assert versions[1]['snapshots'] == ['previous']
        assert versions[1]['version'] == 1
        assert_frame_equal_(library.read(symbol, as_of='previous').data, ts1)

        # ensure snapshot of previous version doesn't overwrite current version
        library.write(symbol, ts1, prune_previous_version=True)
        library.snapshot('another', versions={symbol: 1})
        versions = library.list_versions(symbol)

        assert versions[0]['snapshots'] == []
        assert versions[0]['version'] == 3
        assert_frame_equal_(library.read(symbol).data, ts1)

        assert versions[1]['snapshots'] == ['new']
        assert versions[1]['version'] == 2

        assert versions[2]['snapshots'] == ['previous', 'another']
        assert versions[2]['version'] == 1
        assert_frame_equal_(library.read(symbol, as_of='another').data, ts1)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_snapshot_exclusion(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1)
        library.snapshot('current', skip_symbols=[symbol])
        versions = list(library.list_versions(symbol))
        assert len(versions) == 1
        assert versions[0]['snapshots'] == []


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_snapshot_delete(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1)
        library.snapshot('current')
        library.write(symbol, ts2)

        # We have two versions of the symbol
        assert len(list(library.list_versions(symbol))) == 2
        library.delete_snapshot('current')
        # Data no longer referenced by snapshot
        with pytest.raises(NoDataFoundException):
            library.read(symbol, as_of='current')
        # But still accessible through the version
        assert_frame_equal_(library.read(symbol, as_of=1).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=2).data, ts2)

        # Snapshot again
        library.snapshot('current')
        library.write(symbol, ts1)
        assert_frame_equal_(library.read(symbol, as_of='current').data, ts2)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_multiple_snapshots(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1)
        library.snapshot('current')
        library.write(symbol, ts2)
        library.snapshot('current2')

        assert 'current' in library.list_snapshots()
        assert 'current2' in library.list_snapshots()

        assert_frame_equal_(library.read(symbol).data, ts2)
        assert_frame_equal_(library.read(symbol, as_of=1).data, ts1)
        assert_frame_equal_(library.read(symbol, as_of=2).data, ts2)
        assert_frame_equal_(library.read(symbol, as_of='current').data, ts1)
        assert_frame_equal_(library.read(symbol, as_of='current2').data, ts2)

        library.delete_snapshot('current')
        assert_frame_equal_(library.read(symbol, as_of='current2').data, ts2)
        library.delete_snapshot('current2')
        assert len(list(library.list_versions(symbol))) == 2


def test_delete_identical_snapshots(library):
    library.write(symbol, ts1)
    library.snapshot('current1')
    library.snapshot('current2')
    library.snapshot('current3')

    library.delete_snapshot('current3')
    assert_frame_equal_(library.read(symbol, as_of='current2').data, ts1)
    library.delete_snapshot('current1')
    assert_frame_equal_(library.read(symbol, as_of='current2').data, ts1)
    assert_frame_equal_(library.read(symbol).data, ts1)


def test_list_snapshots(library):
    library.write(symbol, ts1)
    library.snapshot('current')
    library.snapshot('current2')

    assert 'current' in library.list_snapshots()
    assert 'current2' in library.list_snapshots()


def test_duplicate_snapshots(library):
    library.write(symbol, ts1)
    library.snapshot('current')
    with pytest.raises(DuplicateSnapshotException):
        library.snapshot('current')


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prunes_multiple_versions(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        coll = library._collection

        a = [{'a': 'b'}]
        c = [{'c': 'd'}]
        # Create an ObjectId
        now = dt.utcnow()
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=125))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=122))):
            library.write(symbol, c, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=121))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=119))):
            library.write(symbol, c, prune_previous_version=False)
        assert mongo_count(coll.versions) == 4

        # Prunes all versions older than the most recent version that's older than 10 mins
        library.write(symbol, a, prune_previous_version=True)
        assert mongo_count(coll.versions) == 3
        assert library.read(symbol, as_of=3).data == a
        assert library.read(symbol, as_of=4).data == c
        assert library.read(symbol, as_of=5).data == a


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prunes_doesnt_prune_snapshots(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        coll = library._collection

        a = [{'a': 'b'}]
        c = [{'c': 'd'}]
        now = dt.utcnow()
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=125))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=122))):
            library.write(symbol, c, prune_previous_version=False)
        library.snapshot('snap')
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=121))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=119))):
            library.write(symbol, c, prune_previous_version=False)
        assert mongo_count(coll.versions) == 4

        # Prunes all versions older than the most recent version that's older than 10 mins
        library.write(symbol, a, prune_previous_version=True)
        assert mongo_count(coll.versions) == 4
        assert library.read(symbol, as_of='snap').data == c
        assert library.read(symbol, as_of=3).data == a
        assert library.read(symbol, as_of=4).data == c
        assert library.read(symbol, as_of=5).data == a

        # Remove the snapshot, the version should now be pruned
        library.delete_snapshot('snap')
        assert mongo_count(coll.versions) == 4
        library.write(symbol, c, prune_previous_version=True)
        assert mongo_count(coll.versions) == 4
        assert library.read(symbol, as_of=4).data == c
        assert library.read(symbol, as_of=5).data == a
        assert library.read(symbol, as_of=6).data == c


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prunes_multiple_versions_ts(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        coll = library._collection
        a = ts1
        c = ts2
        # Create an ObjectId
        now = dt.utcnow()
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=125))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=122))):
            library.write(symbol, c, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=121))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=119))):
            library.write(symbol, c, prune_previous_version=False)
        assert mongo_count(coll.versions) == 4

        # Prunes all versions older than the most recent version that's older than 10 mins
        library.write(symbol, a, prune_previous_version=True)
        assert mongo_count(coll.versions) == 3
        assert_frame_equal_(library.read(symbol, as_of=3).data, a)
        assert_frame_equal(library.read(symbol, as_of=4).data, c)
        assert_frame_equal_(library.read(symbol, as_of=5).data, a)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prunes_doesnt_prune_snapshots_ts(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        coll = library._collection

        a = ts1
        c = ts2
        now = dt.utcnow()
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=125))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=122))):
            library.write(symbol, c, prune_previous_version=False)
        library.snapshot('snap')
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=121))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=119))):
            library.write(symbol, c, prune_previous_version=False)
        assert mongo_count(coll.versions) == 4

        # Prunes all versions older than the most recent version that's older than 10 mins
        library.write(symbol, a, prune_previous_version=True)
        assert mongo_count(coll.versions) == 4
        assert_frame_equal_(library.read(symbol, as_of='snap').data, c)
        assert_frame_equal_(library.read(symbol, as_of=3).data, a)
        assert_frame_equal_(library.read(symbol, as_of=4).data, c)
        assert_frame_equal_(library.read(symbol, as_of=5).data, a)

        # Remove the snapshot, the version should now be pruned
        library.delete_snapshot('snap')
        assert mongo_count(coll.versions) == 4
        library.write(symbol, c, prune_previous_version=True)
        assert mongo_count(coll.versions) == 4
        assert_frame_equal_(library.read(symbol, as_of=4).data, c)
        assert_frame_equal_(library.read(symbol, as_of=5).data, a)
        assert_frame_equal_(library.read(symbol, as_of=6).data, c)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prunes_multiple_versions_fully_different_tss(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        coll = library._collection

        a = ts1
        b = ts2
        c = b.copy()
        c.index = [i + dtd(days=365) for i in c.index]
        c.index.name = b.index.name
        # Create an ObjectId
        now = dt.utcnow()
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=125))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=124))):
            library.write(symbol, b, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=122))):
            library.write(symbol, c, prune_previous_version=False)
        # a b and c versions above will be pruned a and b share months
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=121))):
            library.write(symbol, c, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=119))):
            library.write(symbol, c, prune_previous_version=False)
        assert mongo_count(coll.versions) == 5

        # Prunes all versions older than the most recent version that's older than 10 mins
        library.write(symbol, c, prune_previous_version=True)
        assert_frame_equal_(library.read(symbol, as_of=4).data, c)
        assert_frame_equal_(library.read(symbol, as_of=5).data, c)
        assert_frame_equal_(library.read(symbol, as_of=6).data, c)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prunes_doesnt_prune_snapshots_fully_different_tss(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        coll = library._collection

        a = ts1
        b = ts2
        c = b.copy()
        c.index = [i + dtd(days=365) for i in c.index]
        c.index.name = b.index.name
        now = dt.utcnow()
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=125))):
            library.write(symbol, a, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=123))):
            library.write(symbol, b, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=122))):
            library.write(symbol, c, prune_previous_version=False)
        library.snapshot('snap')
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=121))):
            library.write(symbol, c, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=118))):
            library.write(symbol, c, prune_previous_version=False)
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=119))):
            library.write(symbol, c, prune_previous_version=False)
        assert mongo_count(coll.versions) == 6

        # Prunes all versions older than the most recent version that's older than 10 mins
        library.write(symbol, c, prune_previous_version=True)
        assert mongo_count(coll.versions) == 5
        assert_frame_equal_(library.read(symbol, as_of='snap').data, c)
        assert_frame_equal_(library.read(symbol, as_of=4).data, c)
        assert_frame_equal_(library.read(symbol, as_of=5).data, c)
        assert_frame_equal_(library.read(symbol, as_of=6).data, c)
        assert_frame_equal_(library.read(symbol, as_of=7).data, c)

        library.delete_snapshot('snap')
        assert mongo_count(coll.versions) == 5
        library.write(symbol, c, prune_previous_version=True)
        assert_frame_equal_(library.read(symbol, as_of=4).data, c)
        assert_frame_equal_(library.read(symbol, as_of=5).data, c)
        assert_frame_equal_(library.read(symbol, as_of=6).data, c)
        assert_frame_equal_(library.read(symbol, as_of=7).data, c)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prunes_previous_version_append_interaction(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        ts = ts1
        ts2 = ts1.append(pd.DataFrame(index=[ts.index[-1] + dtd(days=1),
                                             ts.index[-1] + dtd(days=2), ],
                                      data=[3.7, 3.8],
                                      columns=['near']))
        ts2.index.name = ts1.index.name
        ts3 = ts.append(pd.DataFrame(index=[ts2.index[-1] + dtd(days=1),
                                            ts2.index[-1] + dtd(days=2)],
                                     data=[4.8, 4.9],
                                     columns=['near']))
        ts3.index.name = ts1.index.name
        ts4 = ts
        ts5 = ts2
        ts6 = ts3
        now = dt.utcnow()
        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=130)),
                                    from_datetime=bson.ObjectId.from_datetime):
            library.write(symbol, ts, prune_previous_version=False)
        assert_frame_equal_(ts, library.read(symbol).data)

        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=129)),
                                    from_datetime=bson.ObjectId.from_datetime):
            library.write(symbol, ts2, prune_previous_version=False)
        assert_frame_equal_(ts, library.read(symbol, as_of=1).data)
        assert_frame_equal_(ts2, library.read(symbol).data)

        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=128)),
                                    from_datetime=bson.ObjectId.from_datetime):
            library.write(symbol, ts3, prune_previous_version=False)
        assert_frame_equal_(ts, library.read(symbol, as_of=1).data)
        assert_frame_equal_(ts2, library.read(symbol, as_of=2).data)
        assert_frame_equal_(ts3, library.read(symbol).data)

        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=127)),
                                    from_datetime=bson.ObjectId.from_datetime):
            library.write(symbol, ts4, prune_previous_version=False)
        assert_frame_equal_(ts, library.read(symbol, as_of=1).data)
        assert_frame_equal_(ts2, library.read(symbol, as_of=2).data)
        assert_frame_equal_(ts3, library.read(symbol, as_of=3).data)
        assert_frame_equal_(ts4, library.read(symbol).data)

        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now - dtd(minutes=126)),
                                    from_datetime=bson.ObjectId.from_datetime):
            library.write(symbol, ts5, prune_previous_version=False)
        assert_frame_equal_(ts, library.read(symbol, as_of=1).data)
        assert_frame_equal_(ts2, library.read(symbol, as_of=2).data)
        assert_frame_equal_(ts3, library.read(symbol, as_of=3).data)
        assert_frame_equal_(ts4, library.read(symbol, as_of=4).data)
        assert_frame_equal_(ts5, library.read(symbol).data)

        with patch("bson.ObjectId", return_value=bson.ObjectId.from_datetime(now),
                                    from_datetime=bson.ObjectId.from_datetime):
            library.write(symbol, ts6, prune_previous_version=True)

        with pytest.raises(NoDataFoundException):
            library.read(symbol, as_of=1)
        with pytest.raises(NoDataFoundException):
            library.read(symbol, as_of=2)
        with pytest.raises(NoDataFoundException):
            library.read(symbol, as_of=3)
        assert_frame_equal_(ts5, library.read(symbol, as_of=5).data)
        assert_frame_equal_(ts6, library.read(symbol).data)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_list_symbols(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.snapshot('snap1')
        library.write('asdf', {'foo': 'bar'}, metadata={'a': 1, 'b': 10})
        library.snapshot('snap2')
        assert 'asdf' in library.list_symbols()
        assert 'asdf' not in library.list_symbols(snapshot='snap1')
        assert 'asdf' in library.list_symbols(snapshot='snap2')
        assert 'asdf' in library.list_symbols(all_symbols=True)
        assert 'asdf' in library.list_symbols(a=1)
        assert library.list_symbols(a={'$gt': 5}) == []
        assert library.list_symbols(b={'$gt': 5}) == ['asdf']


def test_snapshot_no_collscan(library):
    with enable_profiling_for_library(library) as profile_db:
        library.write('asdf', {'foo': 'bar'}, metadata={'a': 1, 'b': 10})
        library.snapshot('snap1')

    query_plan = list(profile_db.find({'command.aggregate': {'$exists': True}}))[-1]

    # TODO this can be 0 or 1 depending of query plan being IXSCAN or DEPENDS_SCAN
    assert query_plan.get('docsExamined') <= 1
    assert query_plan.get('keysExamined') == 1
    assert not query_plan.get('planSummary').startswith('COLLSCAN')


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_list_symbols_regex(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.snapshot('snap1')
        library.write('asdf', {'foo': 'bar'}, metadata={'a': 1, 'b': 10})
        library.write('furble', {'foo': 'bar'}, metadata={'a': 1, 'b': 10})
        library.snapshot('snap2')
        assert 'asdf' in library.list_symbols(regex='asd')
        assert 'furble' not in library.list_symbols(regex='asd')
        assert 'asdf' not in library.list_symbols(snapshot='snap1', regex='asd')
        assert 'asdf' in library.list_symbols(snapshot='snap2', regex='asd')
        assert 'furble' not in library.list_symbols(snapshot='snap2', regex='asd')
        assert 'asdf' in library.list_symbols(all_symbols=True, regex='asd')
        assert 'furble' not in library.list_symbols(all_symbols=True, regex='asd')
        assert 'asdf' in library.list_symbols(a=1, regex='asd')
        assert 'furble' not in library.list_symbols(a=1, regex='asd')
        assert library.list_symbols(a={'$gt': 5}, regex='asd') == []
        assert library.list_symbols(b={'$gt': 5}, regex='asd') == ['asdf']


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_list_symbols_newer_version_with_lower_id(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        now = struct.pack(">i", int(time.time()))
        old_id = bson.ObjectId(now + b"\x00\x00\x00\x00\x00\x00\x00\x00")
        new_id = bson.ObjectId(now + b"\x00\x00\x00\x00\x00\x00\x00\x01")
        object_id_class = Mock()
        object_id_class.from_datetime = bson.ObjectId.from_datetime

        object_id_class.return_value = new_id
        with patch("bson.ObjectId", object_id_class):
            library.write(symbol, ts1)

        library.snapshot('s1')

        object_id_class.return_value = old_id
        with patch("bson.ObjectId", object_id_class):
            library.delete(symbol)

        assert symbol not in library.list_symbols()


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_list_symbols_write_snapshot_write_delete(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write('asdf', {'foo': 'bar'})
        symbol = 'sym_a'
        library.write(symbol, {'foo2': 'bar2'}, prune_previous_version=False)
        library.snapshot('s1')
        library.write(symbol, {'foo3': 'bar2'}, prune_previous_version=False)
        library.delete(symbol)
        # at this point we have one version retained from 's1' and
        # one more version added with delete (version with foo3 is pruned)
        # The list_symbols should return only 'asdf'
        assert library.list_symbols() == ['asdf']


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_list_symbols_delete_write(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'sym_a'
        library.write(symbol, {'foo': 'bar2'}, prune_previous_version=False)
        library.delete(symbol)
        library.write(symbol, {'foo2': 'bar2'}, prune_previous_version=False)
        assert library.list_symbols() == [symbol]


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_date_range_large(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        index = [dt(2017, 1, 1)]*20000 + [dt(2017, 1, 2)]*20000
        data = np.random.random((40000, 10))
        df = pd.DataFrame(index=index, data=data)
        df.index.name = 'index'
        df.columns = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']
        library.write('test', df)
        r = library.read('test', date_range=DateRange(dt(2017, 1, 1), dt(2017, 1, 2)))
        assert_frame_equal_(df, r.data)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_append_after_empty(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        len_df = 500
        index = [dt(2017, 1, 2)] * len_df
        data = np.random.random((len_df, 10))
        df = pd.DataFrame(index=index, data=data)
        df.index.name = 'index'
        df.columns = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']
        library.write(symbol, df.iloc[:0])

        for i in range(len_df):
            library.append(symbol, df.iloc[i:i + 1])
        r = library.read(symbol)
        assert_frame_equal_(df, r.data)


def _rnd_df(nrows, ncols):
    ret_df = pd.DataFrame(np.random.randn(nrows, ncols),
                          index=pd.date_range('20170101',
                          periods=nrows, freq='S'),
                          columns=[chr(i) for i in range(ord('a'), ord('a')+ncols)])
    ret_df.index.name = 'index'
    return ret_df


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_metadata(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a, mydf_b = _rnd_df(10, 5), _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write(symbol, data=mydf_b, metadata={'field_a': 2})  # creates version 2
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 3 (only metadata)

            v_doc_2 = library._versions.find_one({'symbol': symbol, 'version': 2})
            v_doc_3 = library._versions.find_one({'symbol': symbol, 'version': 3})
            assert get_fwptr_config(v_doc_2) is fw_pointers_cfg
            assert v_doc_2.get(FW_POINTERS_REFS_KEY) == v_doc_3.get(FW_POINTERS_REFS_KEY)

            v = library.read(symbol)
            assert_frame_equal_(v.data, mydf_b, check_freq=False)
            assert v.metadata == {'field_b': 1}
            assert library._read_metadata(symbol).get('version') == 3
            assert_frame_equal_(library.read(symbol, as_of=1).data, mydf_a, check_freq=False)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_metadata_followed_by_append(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a, mydf_b = _rnd_df(10, 5), _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 2 (only metadata)
            library.append(symbol, data=mydf_b, metadata={'field_c': 1})  # creates version 3

            # Trigger GC now
            time.sleep(2)
            library._prune_previous_versions(symbol, 0)

            v = library.read(symbol)
            assert_frame_equal_(v.data, mydf_a.append(mydf_b), check_freq=False)
            assert v.metadata == {'field_c': 1}
            assert library._read_metadata(symbol).get('version') == 3
            assert_frame_equal_(library.read(symbol, as_of=1).data, mydf_a, check_freq=False)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_metadata_new_symbol(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        with patch('arctic.arctic.logger.info') as info:
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 1 (only metadata)
            v = library.read(symbol)
            assert v.data == None
            assert v.metadata == {'field_b': 1}
            assert library._read_metadata(symbol).get('version') == 1


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_metadata_after_append(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a, mydf_b = _rnd_df(10, 5), _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.append(symbol, data=mydf_b, metadata={'field_a': 2})  # creates version 2
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 3
            v = library.read(symbol)
            assert_frame_equal_(v.data, mydf_a.append(mydf_b), check_freq=False)
            assert v.metadata == {'field_b': 1}
            assert library._read_metadata(symbol).get('version') == 3


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_metadata_purge_previous_versions(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a, mydf_b, mydf_c = _rnd_df(10, 5), _rnd_df(10, 5), _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            with FwPointersCtx(fw_pointers_cfg):
                library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
                library.write(symbol, data=mydf_b, metadata={'field_a': 2})  # creates version 2
                assert library._read_metadata(symbol).get('version') == 2
                library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 3 (only metadata)

                # Trigger GC now
                library._prune_previous_versions(symbol, 0)
                time.sleep(2)

                # Assert the data
                v = library.read(symbol)
                assert_frame_equal_(v.data, mydf_b, check_freq=False)
                assert v.metadata == {'field_b': 1}

                # Check if after snapshot and deleting the symbol, the data/metadata survive
                library.snapshot('SNAP_1')
                library.delete(symbol)
                v = library.read(symbol, as_of='SNAP_1')
                assert_frame_equal_(v.data, mydf_b, check_freq=False)
                assert library._read_metadata(symbol, as_of='SNAP_1').get('version') == 3
                assert v.metadata == {'field_b': 1}


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_metadata_delete_symbol(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        mydf_b = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 2 (only metadata)

            library.delete(symbol)

            with pytest.raises(NoDataFoundException):
                library.read(symbol)

            library.write(symbol, data=mydf_b, metadata={'field_a': 1})  # creates version 1
            assert_frame_equal_(library.read(symbol).data, mydf_b, check_freq=False)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_metadata_snapshots(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a, mydf_b = _rnd_df(10, 5), _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.snapshot('SNAP_1')
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 2 (only metadata)
            library.snapshot('SNAP_2')
            library.write(symbol, data=mydf_b, metadata={'field_c': 1})  # creates version 3
            library.snapshot('SNAP_3')

            library._prune_previous_versions(symbol, keep_mins=0)

            v = library.read(symbol)
            assert_frame_equal_(v.data, mydf_b, check_freq=False)
            assert v.metadata == {'field_c': 1}

            v = library.read(symbol, as_of='SNAP_1')
            assert_frame_equal_(v.data, mydf_a, check_freq=False)
            assert v.metadata == {'field_a': 1}

            v = library.read(symbol, as_of='SNAP_2')
            assert_frame_equal_(v.data, mydf_a, check_freq=False)
            assert v.metadata == {'field_b': 1}


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_version(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        mydf_b = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write(symbol, data=mydf_b, metadata={'field_a': 2})  # creates version 2

            item = library.read(symbol)
            assert_frame_equal_(item.data, mydf_b, check_freq=False)
            assert item.metadata == {'field_a': 2}
            assert library._read_metadata(symbol).get('version') == 2

            restore_item = library.restore_version(symbol, as_of=1)  # creates version 3
            assert restore_item.version == 3
            assert restore_item.metadata == {'field_a': 1}

            item = library.read(symbol)
            assert_frame_equal_(item.data, mydf_a, check_freq=False)
            assert item.metadata == {'field_a': 1}
            assert library._read_metadata(symbol).get('version') == 3


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_version_followed_by_append(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        mydf_b = _rnd_df(10, 5)
        mydf_c = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write(symbol, data=mydf_b, metadata={'field_b': 2})  # creates version 2

            restore_item = library.restore_version(symbol, as_of=1)  # creates version 3
            assert restore_item.version == 3
            assert restore_item.metadata == {'field_a': 1}

            library.append(symbol, data=mydf_c, metadata={'field_c': 3})  # creates version 4

            # Trigger GC now
            library._prune_previous_versions(symbol, 0)
            time.sleep(2)

            item = library.read(symbol)
            assert_frame_equal_(item.data, mydf_a.append(mydf_c), check_freq=False)
            assert item.metadata == {'field_c': 3}
            assert library._read_metadata(symbol).get('version') == 4


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_version_purging_previous_versions(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        mydf_b = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write(symbol, data=mydf_b, metadata={'field_a': 2})  # creates version 2

            restore_item = library.restore_version(symbol, as_of=1)  # creates version 3
            assert restore_item.version == 3
            assert restore_item.metadata == {'field_a': 1}

            # Trigger GC now
            library._prune_previous_versions(symbol, 0)
            time.sleep(2)

            # library._delete_version(symbol, 1)  # delete the original version to test further the robustness/dependency

            item = library.read(symbol)
            assert_frame_equal_(item.data, mydf_a, False)
            assert item.metadata == {'field_a': 1}
            assert library._read_metadata(symbol).get('version') == 3


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_version_non_existent_version(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1

            with pytest.raises(NoDataFoundException):
                library.restore_version(symbol, as_of=3)

            item = library.read(symbol)
            assert_frame_equal_(item.data, mydf_a, check_freq=False)
            assert item.metadata == {'field_a': 1}
            assert item.version == 1


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_version_which_updated_only_metadata(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        mydf_b = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 2
            library.write(symbol, data=mydf_b)  # creates version 3

            restore_item = library.restore_version(symbol, as_of=2)  # creates version 4
            assert restore_item.version == 4
            assert restore_item.metadata == {'field_b': 1}

            item = library.read(symbol)
            assert_frame_equal_(item.data, mydf_a, check_freq=False)
            assert item.metadata == {'field_b': 1}
            assert item.version == 4


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_version_then_snapshot(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        mydf_b = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 2

            restore_item = library.restore_version(symbol, as_of=1)  # creates version 3
            assert restore_item.metadata == {'field_a': 1}
            assert restore_item.version == 3

            library.snapshot('SNAP_1')
            library.write(symbol, data=mydf_b)  # creates version 3

            item = library.read(symbol, as_of='SNAP_1')
            assert_frame_equal_(item.data, mydf_a, check_freq=False)
            assert item.metadata == {'field_a': 1}
            assert item.version == 3


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_version_latest_snapshot_noop(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 2
            library.snapshot('SNAP_1')

            restore_item = library.restore_version(symbol, as_of='SNAP_1')  # does not create a new version
            assert restore_item.metadata == {'field_b': 1}
            assert restore_item.version == 2

            item = library.read(symbol)
            assert_frame_equal_(item.data, mydf_a, check_freq=False)
            assert item.metadata == {'field_b': 1}
            assert item.version == 2


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_version_latest_version_noop(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.write_metadata(symbol, metadata={'field_b': 1})  # creates version 2

            restore_item = library.restore_version(symbol, as_of=2)  # does not create a new version
            assert restore_item.metadata == {'field_b': 1}
            assert restore_item.version == 2

            item = library.read(symbol)
            assert_frame_equal_(item.data, mydf_a, check_freq=False)
            assert item.metadata == {'field_b': 1}
            assert item.version == 2


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_version_snap_delete_symbol_restore(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf = _rnd_df(20, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf[:10], metadata={'field_a': 1})  # creates version 1
            library.append(symbol, data=mydf[10:15])  # version 2
            library.snapshot('snapA')

            library.append(symbol, data=mydf[15:20])  # version 3
            library.delete(symbol)  # version 4

            restored_item = library.restore_version(symbol, as_of='snapA')  # version 5
            assert restored_item.metadata == {'field_a': 1}
            assert restored_item.version == 5

            item = library.read(symbol)
            assert_frame_equal_(item.data, mydf[:15], check_freq=False)
            assert item.metadata == {'field_a': 1}
            assert item.version == 5


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_restore_from_version_with_deleted_symbol(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        symbol = 'FTL'
        mydf_a = _rnd_df(10, 5)
        with patch('arctic.arctic.logger.info') as info:
            library.write(symbol, data=mydf_a, metadata={'field_a': 1})  # creates version 1
            library.delete(symbol)

            with pytest.raises(NoDataFoundException):
                library.restore_version(symbol, as_of=2)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prune_previous_versions_retries_on_cleanup_error(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        original_cleanup = _version_store_utils.cleanup
        def _cleanup(*args, **kwargs):
            if _cleanup.first_try:
                _cleanup.first_try = False
                raise OperationFailure(0)
            else:
                return original_cleanup(*args, **kwargs)
        _cleanup.first_try = True

        library.write(symbol, ts1)
        library.write(symbol, ts2)

        with patch("arctic.store.version_store.cleanup", side_effect=_cleanup) as cleanup:
            cleanup.__name__ = "cleanup"  # required by functools.wraps
            library._prune_previous_versions(symbol, keep_mins=0)

        assert len(list(library._arctic_lib.get_top_level_collection().find({'symbol': symbol}))) == 1


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prune_previous_versions_retries_find_calls(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        original_next = pymongo.cursor.Cursor.next

        callers = set()
        def _next(*args, **kwargs):
            vs_caller_name = next(c for c in inspect.stack() if c[1].endswith('arctic/store/version_store.py'))[3]
            if vs_caller_name not in callers:
                callers.add(vs_caller_name)
                raise OperationFailure(0)
            else:
                return original_next(*args, **kwargs)

        library.write(symbol, ts1, prune_previous_version=False)
        library.write(symbol, ts2, prune_previous_version=False)

        with patch.object(pymongo.cursor.Cursor, "next", autospec=True, side_effect=_next):
            library._prune_previous_versions(symbol, keep_mins=0)

        assert mongo_count(library._versions, filter={'symbol': symbol}) == 1


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_append_does_not_duplicate_data_when_prune_fails(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        side_effect = [OperationFailure(0), arctic.store.version_store.VersionStore._prune_previous_versions]
        new_data = read_str_as_pandas("""times | near
        2013-01-01 17:06:11.040 |  7.0
        2013-01-02 17:06:11.040 |  8.2
        2013-01-03 17:06:11.040 |  3.5
        2013-01-04 17:06:11.040 |  0.7""")
        library.write(symbol, ts1)

        with patch.object(arctic.store.version_store.VersionStore, "_prune_previous_versions", autospec=True, side_effect=side_effect):
            library.append(symbol, new_data)

        data = library.read(symbol).data
        assert len(set(data.index)) == len(data.index)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_does_not_succeed_with_a_prune_error(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        # More than max retries OperationFailure would be more realistic, but ValueError is used for simplicity
        side_effect = [ValueError, arctic.store.version_store.VersionStore._prune_previous_versions]
        library.write(symbol, ts1)

        with patch.object(arctic.store.version_store.VersionStore, "_prune_previous_versions", autospec=True, side_effect=side_effect):
            with pytest.raises(ValueError):
                library.write(symbol, ts1)

        assert len(library.list_versions(symbol)) == 1


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_prune_keeps_version(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write(symbol, ts1)
        library.write(symbol, ts1)
        old_version = [v["_id"] for v in library._versions.find({"symbol": symbol}, sort=[("_id", 1)])][0]

        library._prune_previous_versions(symbol, keep_mins=0, keep_version=old_version)

        assert len(library.list_versions(symbol)) == 2


def test_empty_string_column_name(library):
    df = pd.DataFrame(data=[0, 1, 2], index=[0, 1, 2])
    df.columns = ['']

    with pytest.raises(ArcticException):
        library.write('df', df)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_snapshot_list_versions_after_delete(library, library_name, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        library.write("symA", 'data data')
        library.write("symB", 'data data')
        library.write("symC", 'data data')
        library.snapshot('snapA')

        library.delete('symC')

        assert {v['symbol'] for v in library.list_versions(snapshot='snapA')} == {'symA', 'symB', 'symC'}


def test_write_non_serializable_throws(arctic):
    lib_name = 'write_hanlder_test'
    arctic.initialize_library(lib_name, VERSION_STORE)
    with patch('arctic.store.version_store.STRICT_WRITE_HANDLER_MATCH', True):
        library = arctic[lib_name]

        # Check that falling back to a pickle from a dataframe throws
        df = pd.DataFrame({'a': [dict(a=1)]})

        with pytest.raises(ArcticException):
            library.write('ns1', df)

        # Check that saving a regular dataframe succeeds with this option set
        library.write('ns2', ts1)
        assert_frame_equal_(ts1, library.read('ns2').data, check_freq=False)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_non_serializable_pickling_default(arctic, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        lib_name = 'write_hanlder_test'
        arctic.initialize_library(lib_name, VERSION_STORE)
        library = arctic[lib_name]
        df = pd.DataFrame({'a': [dict(a=1)]})
        library.write('ns3', df)
        assert_frame_equal_(df, library.read('ns3').data, check_freq=False)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_write_strict_no_daterange(arctic, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        lib_name = 'write_hanlder_test'
        arctic.initialize_library(lib_name, VERSION_STORE)

        # Write with pickling
        with patch('arctic.store.version_store.STRICT_WRITE_HANDLER_MATCH', True):
            library = arctic[lib_name]
            data = [dict(a=1)]
            library.write('ns4', data)

            # When the option is set, we should now be unable to read this item when we specify a
            # date range, even though it was written successfully
            with pytest.raises(ArcticException):
                library.read('ns4', date_range=DateRange(dt(2017, 1, 1), dt(2017, 1, 2)))

            assert data == library.read('ns4').data


def test_handler_check_default_false(arctic):
    lib_name = 'write_hanlder_test1'
    arctic.initialize_library(lib_name, VERSION_STORE)
    assert arctic[lib_name]._with_strict_handler_match is False


def test_handler_check_default_osenviron(arctic):
    with patch('arctic.store.version_store.STRICT_WRITE_HANDLER_MATCH', True):
        lib_name = 'write_hanlder_test2'
        arctic.initialize_library(lib_name, VERSION_STORE)
        assert arctic[lib_name]._with_strict_handler_match is True


def test_handler_check_set_false(arctic):
    lib_name = 'write_hanlder_test3'
    arctic.initialize_library(lib_name, VERSION_STORE, strict_write_handler=False)
    assert arctic[lib_name]._with_strict_handler_match is False


def test_handler_check_set_true(arctic):
    lib_name = 'write_hanlder_test4'
    arctic.initialize_library(lib_name, VERSION_STORE, strict_write_handler=True)
    assert arctic[lib_name]._with_strict_handler_match is True


def test_write_df_with_objects_in_index(library):
    df = _mixed_test_data()['multiindex_with_object'][0]
    library.write(symbol='symX', data=df)
    read_data = library.read(symbol='symX').data
    assert_frame_equal_(df, read_data)


def test_write_series_with_objects_in_index(library):
    myseries = _mixed_test_data()['multiindex_with_object'][0]['POSITION']
    library.write(symbol='symX', data=myseries)
    read_data = library.read(symbol='symX').data
    assert_series_equal(myseries, read_data)


@pytest.mark.parametrize('input_series', [_mixed_test_data()['with_some_objects'][0]['n1'],
                                          _mixed_test_data()['multi_column_with_some_objects'][0]['bar']['two']])
def test_write_series_with_some_objects(library, input_series):
    library.write(symbol='symX', data=input_series)
    read_data = library.read(symbol='symX').data
    assert_series_equal(input_series, read_data)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_can_write_tz_aware_data_df(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        mydf = _mixed_test_data()['index_tz_aware'][0]
        library.write(symbol='symTz', data=mydf)
        read_data = library.read(symbol='symTz').data
        # Arctic converts by default the data to UTC, convert back
        read_data.colB = read_data.colB.dt.tz_localize('UTC').dt.tz_convert(read_data.index.tzinfo)
        assert library._versions.find_one({'symbol': 'symTz'})['type'] == PandasDataFrameStore.TYPE
        assert_frame_equal_(mydf, read_data, check_freq=False)


@pytest.mark.parametrize('fw_pointers_cfg', [FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED])
def test_can_write_tz_aware_data_series(library, fw_pointers_cfg):
    with FwPointersCtx(fw_pointers_cfg):
        myseries = _mixed_test_data()['index_tz_aware'][0]['colB']
        library.write(symbol='symTzSer', data=myseries)
        read_data = library.read(symbol='symTzSer').data
        # Arctic converts by default the data to UTC, convert back
        read_data = read_data.dt.tz_localize('UTC').dt.tz_convert(read_data.index.tzinfo)
        assert library._versions.find_one({'symbol': 'symTzSer'})['type'] == PandasSeriesStore.TYPE
        assert_series_equal_(myseries, read_data, check_freq=False)


@pytest.mark.parametrize('write_cfg, read_cfg, append_cfg, reread_cfg', [
    (FwPointersCfg.DISABLED, FwPointersCfg.DISABLED, FwPointersCfg.DISABLED, FwPointersCfg.DISABLED),
    (FwPointersCfg.ENABLED, FwPointersCfg.ENABLED, FwPointersCfg.ENABLED, FwPointersCfg.ENABLED),
    (FwPointersCfg.HYBRID, FwPointersCfg.HYBRID, FwPointersCfg.HYBRID, FwPointersCfg.HYBRID),

    (FwPointersCfg.HYBRID, FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.DISABLED),
    (FwPointersCfg.HYBRID, FwPointersCfg.ENABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED),

    (FwPointersCfg.ENABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED, FwPointersCfg.HYBRID),
    (FwPointersCfg.DISABLED, FwPointersCfg.HYBRID, FwPointersCfg.DISABLED, FwPointersCfg.HYBRID),

    (FwPointersCfg.DISABLED, FwPointersCfg.ENABLED, FwPointersCfg.DISABLED, FwPointersCfg.ENABLED),

    # This throws exception, and ENABLED --> DISABLED fw-pointers config is not allowed/supported
    (FwPointersCfg.ENABLED, FwPointersCfg.DISABLED, FwPointersCfg.ENABLED, FwPointersCfg.DISABLED),
    (FwPointersCfg.ENABLED, FwPointersCfg.ENABLED, FwPointersCfg.DISABLED, FwPointersCfg.DISABLED),
    (FwPointersCfg.ENABLED, FwPointersCfg.ENABLED, FwPointersCfg.HYBRID, FwPointersCfg.DISABLED),
    (FwPointersCfg.ENABLED, FwPointersCfg.ENABLED, FwPointersCfg.HYBRID, FwPointersCfg.ENABLED)
])
def test_fwpointers_mixed_scenarios(library, write_cfg, read_cfg, append_cfg, reread_cfg):
    def _assert_fw_ptr_meta(symbol, fw_cfg):
        last_v = library._versions.find_one({'symbol': symbol}, sort=[("version", pymongo.DESCENDING)])
        if fw_cfg is FwPointersCfg.DISABLED:
            assert FW_POINTERS_REFS_KEY not in last_v
        else:
            assert FW_POINTERS_REFS_KEY in last_v

    def _assert_append(symbol, fw_cfg, prev_v_fw_cfg):
        last_v = library._versions.find_one({'symbol': symbol}, sort=[("version", pymongo.DESCENDING)])
        if fw_cfg is prev_v_fw_cfg:
            assert last_v.get('append_count', 0)
        else:
            # append is converted to write
            assert last_v.get('append_count', 0) == 0

    orig_check_written = arctic.store._ndarray_store.NdarrayStore.check_written
    outer = {'round': -1,     # unfortunately "nonlocal" keyword is not available in Python 2, this is a workaround
             'raised': False}

    def _mock_check_written(self, collection, symbol, version):
        outer['round'] += 1
        if outer['round'] % 2 == 0:
            outer['raised'] = True
            raise pymongo.errors.OperationFailure("Failed to write all the chunks. Mocked failure.")
        orig_check_written(collection, symbol, version)

    symbol = 'sym/{}/{}/{}/{}'.format(write_cfg, read_cfg, append_cfg, reread_cfg)
    mydf = _mixed_test_data()['small'][0]
    to_write = mydf[:len(mydf) // 2]
    to_append = mydf[len(mydf) // 2:]

    with patch('arctic.store._ndarray_store.NdarrayStore.check_written') as mock_check_written:
        mock_check_written.side_effect = _mock_check_written

        # Patch safely the check_written static method to always fail once and trigger mongo_retry for better coverage
        arctic.store._ndarray_store.NdarrayStore.check_written = _mock_check_written

        # The write
        with FwPointersCtx(write_cfg):
            library.write(symbol=symbol, data=to_write)
            _assert_fw_ptr_meta(symbol, write_cfg)

        # The read
        with FwPointersCtx(read_cfg):
            assert_frame_equal_(to_write, library.read(symbol=symbol).data)

        # The append
        with FwPointersCtx(append_cfg):
            library.append(symbol=symbol, data=to_append)
            _assert_fw_ptr_meta(symbol, append_cfg)
            _assert_append(symbol, write_cfg, append_cfg)

        # The final read
        with FwPointersCtx(reread_cfg):
            assert_frame_equal_(mydf, library.read(symbol=symbol).data)

    assert outer['raised']


def test_fwpointers_writemetadata_enabled_disabled(library):
    symbol = 'meta_test'

    mydf = _mixed_test_data()['small'][0]
    to_write_a = mydf[:len(mydf) // 2]
    to_write_b = mydf[len(mydf) // 2:]

    with FwPointersCtx(FwPointersCfg.ENABLED):
        library.write(symbol=symbol, data=to_write_a)
        assert_frame_equal_(to_write_a, library.read(symbol=symbol).data)

    with FwPointersCtx(FwPointersCfg.DISABLED):
        library.write_metadata(symbol=symbol, metadata={'key_a': 100})
        assert library.read(symbol).metadata == {'key_a': 100}

    with FwPointersCtx(FwPointersCfg.ENABLED):
        vitem = library.read(symbol)
        assert vitem.metadata == {'key_a': 100}
        assert_frame_equal_(to_write_a, vitem.data)

    with FwPointersCtx(FwPointersCfg.DISABLED):
        vitem = library.read(symbol)
        assert vitem.metadata == {'key_a': 100}
        assert_frame_equal_(to_write_a, vitem.data)

    with FwPointersCtx(FwPointersCfg.ENABLED):
        library.append(symbol=symbol, data=to_write_b, prune_previous_version=False)
        assert_frame_equal_(mydf, library.read(symbol=symbol).data)

    last_v = library._versions.find_one({'symbol': symbol}, sort=[("version", pymongo.DESCENDING)])
    assert len(last_v.get(FW_POINTERS_REFS_KEY)) == 2

    with FwPointersCtx(FwPointersCfg.ENABLED):
        vitem = library.read(symbol, as_of=2)
        assert vitem.metadata == {'key_a': 100}
        assert_frame_equal_(to_write_a, vitem.data)

    with FwPointersCtx(FwPointersCfg.DISABLED):
        vitem = library.read(symbol, as_of=2)
        assert vitem.metadata == {'key_a': 100}
        assert_frame_equal_(to_write_a, vitem.data)


def test_fwpointer_enabled_write_delete_keep_version_append(library):
    symbol = 'meta_test'

    mydf = _mixed_test_data()['small'][0]
    to_write_a = mydf[:len(mydf) // 2]
    to_write_b = mydf[len(mydf) // 2:]

    with FwPointersCtx(FwPointersCfg.ENABLED):
        library.write(symbol=symbol, data=to_write_a)
        library.append(symbol=symbol, data=to_write_b)
        library.snapshot('snapA')
        assert_frame_equal_(mydf, library.read(symbol=symbol).data)

    with FwPointersCtx(FwPointersCfg.DISABLED):
        library.delete(symbol)
        assert len(list(library._versions.find({'symbol': symbol}))) == 3  # maintain all versions due to snapshot

    with FwPointersCtx(FwPointersCfg.ENABLED):
        library.write(symbol=symbol, data=to_write_a)
        assert_frame_equal_(to_write_a, library.read(symbol=symbol).data)


def test_version_arctic_version(arctic):
    import arctic.store.version_store as vs
    orig_val = vs.ARCTIC_VERSION_NUMERICAL
    try:
        lib_name = 'arctic_version_test'
        arctic.initialize_library(lib_name, VERSION_STORE)
        library = arctic[lib_name]
        for i in range(3):
            vs.ARCTIC_VERSION_NUMERICAL = i
            library.write(symbol=symbol, data="hello world {}".format(i), prune_previous_version=False)
        assert library.get_arctic_version(symbol) == 2
        assert library.get_arctic_version(symbol, as_of=1) == 0
    finally:
        vs.ARCTIC_VERSION_NUMERICAL = orig_val


def test_prune_mixed_fwpointer_configs(library):
    mydf = _mixed_test_data()['large'][0]
    to_write_a = mydf[:len(mydf) // 3]
    to_write_b = mydf[len(mydf) // 3:2*len(mydf) // 3]
    to_write_c = mydf[2*len(mydf) // 3:]

    with FwPointersCtx(FwPointersCfg.DISABLED):
        library.write(symbol, to_write_a)

    with FwPointersCtx(FwPointersCfg.HYBRID):
        library.append(symbol, to_write_b, prune_previous_version=False)

    with FwPointersCtx(FwPointersCfg.ENABLED):
        library.append(symbol, to_write_c, prune_previous_version=False)

    assert_frame_equal_(to_write_a, library.read(symbol, as_of=1).data)
    assert_frame_equal_(pd.concat((to_write_a, to_write_b)), library.read(symbol, as_of=2).data)
    assert_frame_equal_(mydf, library.read(symbol).data)

    library._prune_previous_versions(symbol, keep_mins=0)
    assert mongo_count(library._versions, {'symbol': symbol}) == 1

    assert_frame_equal_(mydf, library.read(symbol).data)


@pytest.mark.parametrize('first_write_cfg, second_write_cfg', [
    (FwPointersCfg.DISABLED, FwPointersCfg.HYBRID),
    (FwPointersCfg.DISABLED, FwPointersCfg.ENABLED),
    (FwPointersCfg.HYBRID, FwPointersCfg.DISABLED),
    (FwPointersCfg.HYBRID, FwPointersCfg.ENABLED),
    (FwPointersCfg.ENABLED, FwPointersCfg.HYBRID),
    (FwPointersCfg.ENABLED, FwPointersCfg.DISABLED),
])
def test_prune_crossref_fwpointer_configs(library, first_write_cfg, second_write_cfg):
    mydf = _mixed_test_data()['large'][0]
    to_write_a = mydf[:len(mydf) // 3]
    to_write_b = mydf[len(mydf) // 3: 2*len(mydf) // 3]
    to_write_c = mydf[2*len(mydf) // 3:]

    to_write_a_b = pd.concat((to_write_a, to_write_b))
    to_write_b_c = pd.concat((to_write_b, to_write_c))

    with FwPointersCtx(first_write_cfg):
        library.write(symbol, to_write_a_b)
    with FwPointersCtx(second_write_cfg):
        library.write(symbol, to_write_b_c, prune_previous_version=False)
    assert_frame_equal_(to_write_b_c, library.read(symbol).data)
    library._prune_previous_versions(symbol, keep_mins=0)
    assert mongo_count(library._versions, {'symbol': symbol}) == 1
    assert_frame_equal_(to_write_b_c, library.read(symbol).data)

    library.delete(symbol)
    assert mongo_count(library._versions, {'symbol': symbol}) == 0
    assert mongo_count(library._collection, {'symbol': symbol}) == 0



import logging
import time
from collections import defaultdict
from threading import RLock

from concurrent.futures import FIRST_COMPLETED

from ._workers_pool import LazySingletonTasksCoordinator
from .async_utils import AsyncRequestType, AsyncRequest
from ..decorators import mongo_retry
from ..exceptions import AsyncArcticException


def _arctic_task_exec(request):
    request.start_time = time.time()
    logging.debug("Executing asynchronous request for {}/{}".format(request.library, request.symbol))
    result = None
    try:
        request.is_running = True
        if request.mongo_retry:
            result = mongo_retry(request.fun)(*request.args, **request.kwargs)
        else:
            result = request.fun(*request.args, **request.kwargs)
    except Exception as e:
        request.exception = e
    finally:
        request.data = result
        request.end_time = time.time()
        request.is_running = False
    return result


class AsyncArctic(LazySingletonTasksCoordinator):
    _instance = None
    _SINGLETON_LOCK = RLock()
    _POOL_LOCK = RLock()

    def __init__(self, pool_size):
        # Only allow creation via get_instance
        if not type(self)._SINGLETON_LOCK._is_owned():
            raise AsyncArcticException("AsyncArctic is a singleton, can't create a new instance")

        # Enforce the singleton pattern
        with type(self)._SINGLETON_LOCK:
            super(AsyncArctic, self).__init__(pool_size)
            self.requests_per_library = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
            self.requests_by_id = dict()
            self.local_shutdown = False
            self.deferred_requests = list()

    def __reduce__(self):
        return "ASYNC_ARCTIC"

    def _get_modifiers(self, library_name, symbol=None):
        return self.requests_per_library[library_name][symbol][AsyncRequestType.MODIFIER]

    def _get_accessors(self, library_name, symbol=None):
        return self.requests_per_library[library_name][symbol][AsyncRequestType.ACCESSOR]

    @staticmethod
    def _verify_request(store, is_modifier, **kwargs):
        library_name = None if store is None else store._arctic_lib.get_name()
        symbol = kwargs.get('symbol')
        kind = AsyncRequestType.MODIFIER if is_modifier else AsyncRequestType.ACCESSOR
        callback = kwargs.get('async_callback')
        mongo_retry = bool(kwargs.get('mongo_retry'))
        return library_name, symbol, kind, callback, mongo_retry

    def _is_clashing(self, request):
        return bool(self._get_modifiers(request.library, request.symbol) or
                    request.kind is AsyncRequestType.MODIFIER and self._get_accessors(request.library, request.symbol))

    def _add_request(self, request):
        self.requests_per_library[request.library][request.symbol][request.kind].append(request)
        self.requests_by_id[request.id] = request

    def _remove_request(self, request):
        self.requests_per_library[request.library][request.symbol][request.kind].remove(request)
        if request.id in self.requests_by_id:
            del self.requests_by_id[request.id]

    def _schedule_request(self, request):
        try:
            new_id, new_future = self.submit_task(False, _arctic_task_exec, request)
            request.id = new_id
            # Update the state of tracked tasks
            self._add_request(request)
            request.future = new_future
            request.future.add_done_callback(lambda the_future: self._request_finished(request))
        except Exception:
            # clean up the state
            self._remove_request(request)
            raise

    def submit_arctic_request(self, store, fun, is_modifier, *args, **kwargs):
        lib_name, symbol, kind, callback, mongo_retry = AsyncArctic._verify_request(store, is_modifier, **kwargs)

        for k in ('async_callback', 'mongo_retry'):
            kwargs.pop(k, None)

        with type(self)._POOL_LOCK:  # class level lock, since it is a Singleton
            if self.local_shutdown:
                raise AsyncArcticException("AsyncArctic has been shutdown and can no longer accept new requests.")

            # Create the request object
            request = AsyncRequest(kind, lib_name, fun, callback, *args, **kwargs)

            if lib_name and self._is_clashing(request):
                self.deferred_requests.append(request)
                return request

            self._schedule_request(request)

        return request

    def _reschedule_deferred(self):
        picked = None
        try:
            for deferred in self.deferred_requests:
                if not self._is_clashing(deferred):
                    picked = deferred
                    self._schedule_request(deferred)
                    break
        except:
            logging.exception("Failed to re-schedule a deferred task: {}".format(picked))
            return
        self.deferred_requests.remove(picked)

    def _request_finished(self, request):
        with type(self)._POOL_LOCK:
            self._remove_request(request)
            if self.deferred_requests:
                self._reschedule_deferred()
            elif self.local_shutdown:
                # Deferred shutdown of the underlying pool until the deferred jobs have finished
                super(AsyncArctic, self).shutdown()
        request.is_completed = True
        if callable(request.callback):
            request.callback(request)

    def reset(self, pool_size=None, timeout=None):
        self.shutdown(timeout=timeout)
        self.await_termination(timeout=timeout)
        super(AsyncArctic, self).reset(pool_size, timeout)
        with type(self)._SINGLETON_LOCK:
            self.requests_per_library = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
            self.requests_by_id = dict()
            self.local_shutdown = False
            self.deferred_requests = list()

    def shutdown(self, timeout=None):
        if self.local_shutdown:
            return
        with type(self)._POOL_LOCK:
            self.local_shutdown = True
            if self.total_pending_requests() == 0:
                # safe to use this non-atomic check here, as self.deferred_requests
                # is updated only after the deferred request is scheduled
                super(AsyncArctic, self).shutdown(timeout=timeout)

    def await_termination(self, timeout=None):
        while self.total_pending_requests() > 0:
            AsyncArctic.wait_requests(self.requests_by_id.values() + self.deferred_requests,
                                      do_raise=False, timeout=timeout)
        # super(AsyncArctic, self).await_termination(timeout)

    def total_pending_requests(self):
        with type(self)._POOL_LOCK:  # the lock here is "really" necessary
            return len(self.requests_by_id) + len(self.deferred_requests)

    @staticmethod
    def _wait_until_scheduled(requests, timeout=None, check_interval=0.1):
        start = time.time()
        while True:
            if any(r for r in requests if r.future is None):
                time.sleep(check_interval)
            else:
                return True
            if timeout is not None and time.time() - start >= timeout:
                break
        return False

    @staticmethod
    def wait_request(request, do_raise=False, timeout=None):
        if request is None:
            return
        if not AsyncArctic._wait_until_scheduled((request,), timeout):
            raise AsyncArcticException("Timed-out while waiting for request to be scheduled")
        while request.is_completed:
            AsyncArctic.wait_tasks((request.future,), timeout=timeout, raise_exceptions=do_raise)

    @staticmethod
    def wait_requests(requests, do_raise=False, timeout=None):
        if not AsyncArctic._wait_until_scheduled(requests, timeout):
            raise AsyncArcticException("Timed-out while waiting for request to be scheduled")
        while requests and not all(r.is_completed for r in requests):
            AsyncArctic.wait_tasks(tuple(r.future for r in requests if not r.is_completed and r.future is not None),
                                   timeout=timeout, raise_exceptions=do_raise)

    @staticmethod
    def wait_any_request(requests, do_raise=False, timeout=None):
        if not AsyncArctic._wait_until_scheduled(requests, timeout):
            raise AsyncArcticException("Timed-out while waiting for request to be scheduled")
        while requests and not any(r.is_completed for r in requests):
            AsyncArctic.wait_tasks(tuple(r.future for r in requests if not r.is_completed and r.future is not None),
                                   timeout=timeout, return_when=FIRST_COMPLETED, raise_exceptions=do_raise)

    @staticmethod
    def filter_finished_requests(requests, do_raise=True):
        if not requests:
            return requests, requests
        alive_requests = [r for r in requests if not r.is_completed]
        done_requests = [r for r in requests if r.is_completed]
        if do_raise:
            AsyncArctic.raise_errored(done_requests)
        return alive_requests, done_requests

    @staticmethod
    def raise_first_errored(requests):
        errored = tuple(r for r in requests if r.is_completed and r.exception is not None)
        if errored:
            raise errored[0].exception

    @staticmethod
    def filter_errored(requests):
        return tuple(r for r in requests if r.is_completed and r.exception is not None)


ASYNC_ARCTIC = AsyncArctic.get_instance()
async_arctic_submit = ASYNC_ARCTIC.submit_arctic_request
async_wait_request = ASYNC_ARCTIC.wait_request
async_wait_requests = ASYNC_ARCTIC.wait_requests
async_shutdown = ASYNC_ARCTIC.shutdown
async_await_termination = ASYNC_ARCTIC.await_termination
async_reset_pool = ASYNC_ARCTIC.reset
async_total_requests = ASYNC_ARCTIC.total_pending_requests


# def async_modifier(func):
#     @wraps(func)
#     def wrapper(self, *args, **kwargs):
#         return async_arctic_submit(self, func, True, *args, **kwargs)
#
#     return wrapper
#
#
# def async_accessor(func):
#     @wraps(func)
#     def wrapper(self, *args, **kwargs):
#         return async_arctic_submit(self, func, False, *args, **kwargs)
#
#     return wrapper



from __future__ import print_function

import copy
import logging
from datetime import datetime as dt, timedelta

import numpy as np
import pandas as pd
import pymongo
from bson.binary import Binary
from pymongo import ReadPreference
from pymongo.errors import OperationFailure

try:
    from pandas.core.frame import _arrays_to_mgr
except ImportError:
    # Deprecated since pandas 0.23.4
    from pandas.core.internals.construction import arrays_to_mgr as _arrays_to_mgr

try:
    from pandas.api.types import infer_dtype
except ImportError:
    from pandas.lib import infer_dtype

from ..date import DateRange, to_pandas_closed_closed, mktz, datetime_to_ms, ms_to_datetime, CLOSED_CLOSED, to_dt, utc_dt_to_local_dt
from ..decorators import mongo_retry
from ..exceptions import OverlappingDataException, NoDataFoundException, UnorderedDataException, UnhandledDtypeException, ArcticException
from .._util import indent

try:
    from lz4.block import compress as lz4_compress, decompress as lz4_decompress
    lz4_compressHC = lambda _str: lz4_compress(_str, mode='high_compression')
except ImportError as e:
    from lz4 import compress as lz4_compress, compressHC as lz4_compressHC, decompress as lz4_decompress


PD_VER = pd.__version__
logger = logging.getLogger(__name__)

# Example-Schema:
# --------------
# {ID: ObjectId('52b1d39eed5066ab5e87a56d'),
#  SYMBOL: u'symbol'
#  INDEX: Binary('...', 0),
#  IMAGE_DOC: { IMAGE:  {
#                          'ASK': 10.
#                          ...
#                        }
#              's': <sequence_no>
#              't': DateTime(...)
#             }
#  COLUMNS: {
#   'ACT_FLAG1': {
#        DATA: Binary('...', 0),
#        DTYPE: u'U1',
#        ROWMASK: Binary('...', 0)},
#   'ACVOL_1': {
#        DATA: Binary('...', 0),
#        DTYPE: u'float64',
#        ROWMASK: Binary('...', 0)},
#               ...
#    }
#  START: DateTime(...),
#  END: DateTime(...),
#  END_SEQ: 31553879L,
#  SEGMENT: 1386933906826L,
#  SHA: 1386933906826L,
#  VERSION: 3,
# }

TICK_STORE_TYPE = 'TickStoreV3'

ID = '_id'
SYMBOL = 'sy'
INDEX = 'i'
START = 's'
END = 'e'
START_SEQ = 'sS'
END_SEQ = 'eS'
SEGMENT = 'se'
SHA = 'sh'
IMAGE_DOC = 'im'
IMAGE = 'i'

COLUMNS = 'cs'
DATA = 'd'
DTYPE = 't'
IMAGE_TIME = 't'
ROWMASK = 'm'

COUNT = 'c'
VERSION = 'v'

META = 'md'

CHUNK_VERSION_NUMBER = 3


class TickStore(object):

    @classmethod
    def initialize_library(cls, arctic_lib, **kwargs):
        TickStore(arctic_lib)._ensure_index()

    @mongo_retry
    def _ensure_index(self):
        collection = self._collection
        collection.create_index([(SYMBOL, pymongo.ASCENDING),
                                 (START, pymongo.ASCENDING)], background=True)
        collection.create_index([(START, pymongo.ASCENDING)], background=True)

        self._metadata.create_index([(SYMBOL, pymongo.ASCENDING)], background=True, unique=True)

    def __init__(self, arctic_lib, chunk_size=100000):
        """
        Parameters
        ----------
        arctic_lib : ArcticLibraryBinding
            Arctic Library
        chunk_size : int
            Number of ticks to store in a document before splitting to another document.
            if the library was obtained through get_library then set with: self._chuck_size = 10000
        """
        self._arctic_lib = arctic_lib
        # Do we allow reading from secondaries
        self._allow_secondary = self._arctic_lib.arctic._allow_secondary
        self._chunk_size = chunk_size
        self._reset()

    @mongo_retry
    def _reset(self):
        # The default collections
        self._collection = self._arctic_lib.get_top_level_collection()
        self._metadata = self._collection.metadata

    def __getstate__(self):
        return {'arctic_lib': self._arctic_lib}

    def __setstate__(self, state):
        return TickStore.__init__(self, state['arctic_lib'])

    def __str__(self):
        return """<%s at %s>
%s""" % (self.__class__.__name__, hex(id(self)), indent(str(self._arctic_lib), 4))

    def __repr__(self):
        return str(self)

    def delete(self, symbol, date_range=None):
        """
        Delete all chunks for a symbol.

        Which are, for the moment, fully contained in the passed in
        date_range.

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        date_range : `date.DateRange`
            DateRange to delete ticks in
        """
        query = {SYMBOL: symbol}
        date_range = to_pandas_closed_closed(date_range)
        if date_range is not None:
            assert date_range.start and date_range.end
            query[START] = {'$gte': date_range.start}
            query[END] = {'$lte': date_range.end}
        else:
            # delete metadata on complete deletion
            self._metadata.delete_one({SYMBOL: symbol})
        return self._collection.delete_many(query)

    def list_symbols(self, date_range=None):
        return self._collection.distinct(SYMBOL)

    def _mongo_date_range_query(self, symbol, date_range):
        # Handle date_range
        if not date_range:
            date_range = DateRange()

        # We're assuming CLOSED_CLOSED on these Mongo queries
        assert date_range.interval == CLOSED_CLOSED

        # Since we only index on the start of the chunk,
        # we do a pre-flight aggregate query to find the point where the
        # earliest relevant chunk starts.

        start_range = {}
        first_dt = last_dt = None
        if date_range.start:
            assert date_range.start.tzinfo
            start = date_range.start

            # If all chunks start inside of the range, we default to capping to our
            # range so that we don't fetch any chunks from the beginning of time
            start_range['$gte'] = start

            match = self._symbol_query(symbol)
            match.update({'s': {'$lte': start}})

            result = self._collection.aggregate([
                # Only look at the symbols we are interested in and chunks that
                # start before our start datetime
                {'$match': match},
                # Throw away everything but the start of every chunk and the symbol
                {'$project': {'_id': 0, 's': 1, 'sy': 1}},
                # For every symbol, get the latest chunk start (that is still before
                # our sought start)
                {'$group': {'_id': '$sy', 'start': {'$max': '$s'}}},
                {'$sort': {'start': 1}},
            ])
            # Now we need to get the earliest start of the chunk that still spans the start point.
            # Since we got them sorted by start, we just need to fetch their ends as well and stop
            # when we've seen the first such chunk
            try:
                for candidate in result:
                    chunk = self._collection.find_one({'s': candidate['start'], 'sy': candidate['_id']}, {'e': 1})
                    if chunk['e'].replace(tzinfo=mktz('UTC')) >= start:
                        start_range['$gte'] = candidate['start'].replace(tzinfo=mktz('UTC'))
                        break
            except StopIteration:
                pass

        # Find the end bound
        if date_range.end:
            # If we have an end, we are only interested in the chunks that start before the end.
            assert date_range.end.tzinfo
            last_dt = date_range.end
        else:
            logger.info("No end provided.  Loading a month for: {}:{}".format(symbol, first_dt))
            if not first_dt:
                first_doc = self._collection.find_one(self._symbol_query(symbol),
                                                      projection={START: 1, ID: 0},
                                                      sort=[(START, pymongo.ASCENDING)])
                if not first_doc:
                    raise NoDataFoundException()

                first_dt = first_doc[START]
            last_dt = first_dt + timedelta(days=30)
        if last_dt:
            start_range['$lte'] = last_dt

        # Return chunks in the specified range
        if not start_range:
            return {}
        return {START: start_range}

    def _symbol_query(self, symbol):
        if isinstance(symbol, str):
            query = {SYMBOL: symbol}
        elif symbol is not None:
            query = {SYMBOL: {'$in': symbol}}
        else:
            query = {}
        return query

    def _read_preference(self, allow_secondary):
        """ Return the mongo read preference given an 'allow_secondary' argument
        """
        allow_secondary = self._allow_secondary if allow_secondary is None else allow_secondary
        return ReadPreference.NEAREST if allow_secondary else ReadPreference.PRIMARY

    def read(self, symbol, date_range=None, columns=None, include_images=False, allow_secondary=None,
             _target_tick_count=0):
        """
        Read data for the named symbol.  Returns a VersionedItem object with
        a data and metdata element (as passed into write).

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        date_range : `date.DateRange`
            Returns ticks in the specified DateRange
        columns : `list` of `str`
            Columns (fields) to return from the tickstore
        include_images : `bool`
            Should images (/snapshots) be included in the read
        allow_secondary : `bool` or `None`
            Override the default behavior for allowing reads from secondary members of a cluster:
            `None` : use the settings from the top-level `Arctic` object used to query this version store.
            `True` : allow reads from secondary members
            `False` : only allow reads from primary members

        Returns
        -------
        pandas.DataFrame of data
        """
        perf_start = dt.now()
        rtn = {}
        column_set = set()

        multiple_symbols = not isinstance(symbol, str)

        date_range = to_pandas_closed_closed(date_range)
        query = self._symbol_query(symbol)
        query.update(self._mongo_date_range_query(symbol, date_range))

        if columns:
            projection = dict([(SYMBOL, 1),
                               (INDEX, 1),
                               (START, 1),
                               (VERSION, 1),
                               (IMAGE_DOC, 1)] +
                              [(COLUMNS + '.%s' % c, 1) for c in columns])
            column_set.update([c for c in columns if c != 'SYMBOL'])
        else:
            projection = dict([(SYMBOL, 1),
                               (INDEX, 1),
                               (START, 1),
                               (VERSION, 1),
                               (COLUMNS, 1),
                               (IMAGE_DOC, 1)])

        column_dtypes = {}
        ticks_read = 0
        data_coll = self._collection.with_options(read_preference=self._read_preference(allow_secondary))
        for b in data_coll.find(query, projection=projection).sort([(START, pymongo.ASCENDING)],):
            data = self._read_bucket(b, column_set, column_dtypes,
                                     multiple_symbols or (columns is not None and 'SYMBOL' in columns),
                                     include_images, columns)
            for k, v in data.items():
                try:
                    rtn[k].append(v)
                except KeyError:
                    rtn[k] = [v]
            # For testing
            ticks_read += len(data[INDEX])
            if _target_tick_count and ticks_read > _target_tick_count:
                break

        if not rtn:
            raise NoDataFoundException("No Data found for {} in range: {}".format(symbol, date_range))
        rtn = self._pad_and_fix_dtypes(rtn, column_dtypes)

        index = pd.DatetimeIndex(np.concatenate(rtn[INDEX]).astype('datetime64[ms]'), tz='UTC')
        if columns is None:
            columns = [x for x in rtn.keys() if x not in (INDEX, 'SYMBOL')]
        if multiple_symbols and 'SYMBOL' not in columns:
            columns = ['SYMBOL', ] + columns

        if len(index) > 0:
            arrays = [np.concatenate(rtn[k]) for k in columns]
        else:
            arrays = [[] for _ in columns]

        if multiple_symbols:
            sort = np.argsort(index, kind='mergesort')
            index = index[sort]
            arrays = [a[sort] for a in arrays]

        t = (dt.now() - perf_start).total_seconds()
        logger.info("Got data in %s secs, creating DataFrame..." % t)
        if pd.__version__.startswith("0.") or pd.__version__.startswith("1.0"):
            mgr = _arrays_to_mgr(arrays, columns, index, columns, dtype=None)
        else:
            # 4th argument removed + new argument typ is mandatory
            mgr = _arrays_to_mgr(arrays, columns, index, dtype=None, typ="array")

        rtn = pd.DataFrame(mgr)
        # Present data in the user's default TimeZone
        rtn.index = rtn.index.tz_convert(mktz())

        t = (dt.now() - perf_start).total_seconds()
        ticks = len(rtn)
        rate = int(ticks / t) if t != 0 else float("nan")
        logger.info("%d rows in %s secs: %s ticks/sec" % (ticks, t, rate))
        if not rtn.index.is_monotonic_increasing:
            logger.error("TimeSeries data is out of order, sorting!")
            rtn = rtn.sort_index(kind='mergesort')
        if date_range:
            # FIXME: support DateRange.interval...
            rtn = rtn.loc[date_range.start:date_range.end]
        return rtn

    def read_metadata(self, symbol):
        """
        Read metadata for the specified symbol

        Parameters
        ----------
        symbol : `str`
            symbol name for the item

        Returns
        -------
        dict
        """
        return self._metadata.find_one({SYMBOL: symbol})[META]

    def _pad_and_fix_dtypes(self, cols, column_dtypes):
        # Pad out Nones with empty arrays of appropriate dtypes
        rtn = {}
        index = cols[INDEX]
        full_length = len(index)
        for k, v in cols.items():
            if k != INDEX and k != 'SYMBOL':
                col_len = len(v)
                if col_len < full_length:
                    v = ([None, ] * (full_length - col_len)) + v
                    assert len(v) == full_length
                for i, arr in enumerate(v):
                    if arr is None:
                        #  Replace Nones with appropriate-length empty arrays
                        v[i] = self._empty(len(index[i]), column_dtypes.get(k))
                    else:
                        # Promote to appropriate dtype only if we can safely cast all the values
                        # This avoids the case with strings where None is cast as 'None'.
                        # Casting the object to a string is not worthwhile anyway as Pandas changes the
                        # dtype back to objectS
                        if (i == 0 or v[i].dtype != v[i - 1].dtype) and np.can_cast(v[i].dtype, column_dtypes[k],
                                                                                    casting='safe'):
                            v[i] = v[i].astype(column_dtypes[k], casting='safe')

            rtn[k] = v
        return rtn

    def _set_or_promote_dtype(self, column_dtypes, c, dtype):
        existing_dtype = column_dtypes.get(c)
        if existing_dtype is None or existing_dtype != dtype:
            # Promote ints to floats - as we can't easily represent NaNs
            if np.issubdtype(dtype, int):
                dtype = np.dtype('f8')
            column_dtypes[c] = np.promote_types(column_dtypes.get(c, dtype), dtype)

    def _prepend_image(self, document, im, rtn_length, column_dtypes, column_set, columns):
        image = im[IMAGE]
        first_dt = im[IMAGE_TIME]
        if not first_dt.tzinfo:
            first_dt = first_dt.replace(tzinfo=mktz('UTC'))
        document[INDEX] = np.insert(document[INDEX], 0, np.uint64(datetime_to_ms(first_dt)))
        for field in image:
            if field == INDEX:
                continue
            if columns and field not in columns:
                continue
            if field not in document or document[field] is None:
                col_dtype = np.dtype(str if isinstance(image[field], str) else 'f8')
                document[field] = self._empty(rtn_length, dtype=col_dtype)
                column_dtypes[field] = col_dtype
                column_set.add(field)
            val = image[field]
            document[field] = np.insert(document[field], 0, document[field].dtype.type(val))
        # Now insert rows for fields in document that are not in the image
        for field in set(document).difference(set(image)):
            if field == INDEX:
                continue
            logger.debug("Field %s is missing from image!" % field)
            if document[field] is not None:
                val = np.nan
                document[field] = np.insert(document[field], 0, document[field].dtype.type(val))
        return document

    def _read_bucket(self, doc, column_set, column_dtypes, include_symbol, include_images, columns):
        rtn = {}
        if doc[VERSION] != 3:
            raise ArcticException("Unhandled document version: %s" % doc[VERSION])
        # np.cumsum copies the read-only array created with frombuffer
        rtn[INDEX] = np.cumsum(np.frombuffer(lz4_decompress(doc[INDEX]), dtype='uint64'))
        doc_length = len(rtn[INDEX])
        column_set.update(doc[COLUMNS].keys())

        # get the mask for the columns we're about to load
        union_mask = np.zeros((doc_length + 7) // 8, dtype='uint8')
        for c in column_set:
            try:
                coldata = doc[COLUMNS][c]
                # the or below will make a copy of this read-only array
                mask = np.frombuffer(lz4_decompress(coldata[ROWMASK]), dtype='uint8')
                union_mask = union_mask | mask
            except KeyError:
                rtn[c] = None
        union_mask = np.unpackbits(union_mask)[:doc_length].astype('bool')
        rtn_length = np.sum(union_mask)

        rtn[INDEX] = rtn[INDEX][union_mask]
        if include_symbol:
            rtn['SYMBOL'] = [doc[SYMBOL], ] * rtn_length

        # Unpack each requested column in turn
        for c in column_set:
            try:
                coldata = doc[COLUMNS][c]
                dtype = np.dtype(coldata[DTYPE])
                # values ends up being copied by pandas before being returned to the user. However, we
                # copy it into a bytearray here for safety.
                values = np.frombuffer(bytearray(lz4_decompress(coldata[DATA])), dtype=dtype)
                self._set_or_promote_dtype(column_dtypes, c, dtype)
                rtn[c] = self._empty(rtn_length, dtype=column_dtypes[c])
                # unpackbits will make a copy of the read-only array created by frombuffer
                rowmask = np.unpackbits(np.frombuffer(lz4_decompress(coldata[ROWMASK]),
                                        dtype='uint8'))[:doc_length].astype('bool')
                rowmask = rowmask[union_mask]
                rtn[c][rowmask] = values
            except KeyError:
                rtn[c] = None

        if include_images and doc.get(IMAGE_DOC, {}).get(IMAGE, {}):
            rtn = self._prepend_image(rtn, doc[IMAGE_DOC], rtn_length, column_dtypes, column_set, columns)
        return rtn

    def _empty(self, length, dtype):
        if dtype is not None and dtype == np.float64:
            rtn = np.empty(length, dtype)
            rtn[:] = np.nan
            return rtn
        else:
            return np.empty(length, dtype=np.object_)

    def stats(self):
        """
        Return storage statistics about the library

        Returns
        -------
        dictionary of storage stats
        """
        res = {}
        db = self._collection.database
        conn = db.connection
        res['sharding'] = {}
        try:
            sharding = conn.config.databases.find_one({'_id': db.name})
            if sharding:
                res['sharding'].update(sharding)
            res['sharding']['collections'] = list(conn.config.collections.find(
                                                  {'_id': {'$regex': '^' + db.name + r"\..*"}}))
        except OperationFailure:
            # Access denied
            pass
        res['dbstats'] = db.command('dbstats')
        res['chunks'] = db.command('collstats', self._collection.name)
        res['totals'] = {'count': res['chunks']['count'],
                         'size': res['chunks']['size'],
                         }
        return res

    def _assert_nonoverlapping_data(self, symbol, start, end):
        #
        # Imagine we're trying to insert a tick bucket like:
        #      |S------ New-B -------------- E|
        #  |---- 1 ----| |----- 2 -----| |----- 3 -----|
        #
        # S = New-B Start
        # E = New-B End
        # New-B overlaps with existing buckets 1,2,3
        #
        # All we need to do is find the bucket who's start is immediately before (E)
        # If that document's end is > S, then we know it overlaps
        # with this bucket.
        doc = self._collection.find_one({SYMBOL: symbol,
                                         START: {'$lt': end}
                                         },
                                        projection={START: 1,
                                                    END: 1,
                                                    '_id': 0},
                                        sort=[(START, pymongo.DESCENDING)])
        if doc:
            if not doc[END].tzinfo:
                doc[END] = doc[END].replace(tzinfo=mktz('UTC'))
            if doc[END] > start:
                raise OverlappingDataException(
                    "Document already exists with start:{} end:{} in the range of our start:{} end:{}".format(
                        doc[START], doc[END], start, end))

    def write(self, symbol, data, initial_image=None, metadata=None):
        """
        Writes a list of market data events.

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        data : list of dicts or a pandas.DataFrame
            List of ticks to store to the tick-store.
            if a list of dicts, each dict must contain a 'index' datetime
            if a pandas.DataFrame the index must be a Timestamp that can be converted to a datetime.
            Index names will not be preserved.
        initial_image : dict
            Dict of the initial image at the start of the document. If this contains a 'index' entry it is
            assumed to be the time of the timestamp of the index
        metadata: dict
            optional user defined metadata - one per symbol
        """
        pandas = False
        # Check for overlapping data
        if isinstance(data, list):
            start = data[0]['index']
            end = data[-1]['index']
        elif isinstance(data, pd.DataFrame):
            start = data.index[0].to_pydatetime()
            end = data.index[-1].to_pydatetime()
            pandas = True
        else:
            raise UnhandledDtypeException("Can't persist type %s to tickstore" % type(data))
        self._assert_nonoverlapping_data(symbol, to_dt(start), to_dt(end))

        if pandas:
            buckets = self._pandas_to_buckets(data, symbol, initial_image)
        else:
            buckets = self._to_buckets(data, symbol, initial_image)
        self._write(buckets)

        if metadata:
            self._metadata.replace_one({SYMBOL: symbol},
                                       {SYMBOL: symbol, META: metadata},
                                       upsert=True)

    def _write(self, buckets):
        start = dt.now()
        mongo_retry(self._collection.insert_many)(buckets)
        t = (dt.now() - start).total_seconds()
        ticks = len(buckets) * self._chunk_size
        rate = int(ticks / t) if t != 0 else float("nan")
        logger.debug("%d buckets in %s: approx %s ticks/sec" % (len(buckets), t, rate))

    def _pandas_to_buckets(self, x, symbol, initial_image):
        rtn = []
        for i in range(0, len(x), self._chunk_size):
            bucket, initial_image = TickStore._pandas_to_bucket(x[i:i + self._chunk_size], symbol, initial_image)
            rtn.append(bucket)
        return rtn

    def _to_buckets(self, x, symbol, initial_image):
        rtn = []
        for i in range(0, len(x), self._chunk_size):
            bucket, initial_image = TickStore._to_bucket(x[i:i + self._chunk_size], symbol, initial_image)
            rtn.append(bucket)
        return rtn

    @staticmethod
    def _to_ms(date):
        if isinstance(date, dt):
            if not date.tzinfo:
                logger.warning('WARNING: treating naive datetime as UTC in write path')
            return datetime_to_ms(date)
        return date

    @staticmethod
    def _str_dtype(dtype):
        """
        Represent dtypes without byte order, as earlier Java tickstore code doesn't support explicit byte order.
        """
        assert dtype.byteorder != '>'
        if dtype.kind == 'i':
            assert dtype.itemsize == 8
            return 'int64'
        elif dtype.kind == 'f':
            assert dtype.itemsize == 8
            return 'float64'
        elif dtype.kind == 'U':
            return 'U%d' % (dtype.itemsize / 4)
        else:
            raise UnhandledDtypeException("Bad dtype '%s'" % dtype)

    @staticmethod
    def _ensure_supported_dtypes(array):
        # We only support these types for now, as we need to read them in Java
        if array.dtype.kind == 'i':
            array = array.astype('<i8')
        elif array.dtype.kind == 'f':
            array = array.astype('<f8')
        elif array.dtype.kind in ('O', 'U', 'S'):
            if array.dtype.kind == 'O' and infer_dtype(array) not in ['unicode', 'string', 'bytes']:
                # `string` in python2 and `bytes` in python3
                raise UnhandledDtypeException("Casting object column to string failed")
            try:
                array = array.astype(np.unicode_)
            except (UnicodeDecodeError, SystemError):
                # `UnicodeDecodeError` in python2 and `SystemError` in python3
                array = np.array([s.decode('utf-8') for s in array])
            except:
                raise UnhandledDtypeException("Only unicode and utf8 strings are supported.")
        else:
            raise UnhandledDtypeException("Unsupported dtype '%s' - only int64, float64 and U are supported" % array.dtype)
        # Everything is little endian in tickstore
        if array.dtype.byteorder != '<':
            array = array.astype(array.dtype.newbyteorder('<'))
        return array

    @staticmethod
    def _pandas_compute_final_image(df, image, end):
        # Compute the final image with forward fill of df applied to the image
        final_image = copy.copy(image)
        last_values = df.ffill().tail(1).to_dict()
        last_dict = {i: list(a.values())[0] for i, a in last_values.items()}
        final_image.update(last_dict)
        final_image['index'] = end
        return final_image

    @staticmethod
    def _pandas_to_bucket(df, symbol, initial_image):
        rtn = {SYMBOL: symbol, VERSION: CHUNK_VERSION_NUMBER, COLUMNS: {}, COUNT: len(df)}
        end = to_dt(df.index[-1].to_pydatetime())
        if initial_image:
            if 'index' in initial_image:
                start = min(to_dt(df.index[0].to_pydatetime()), initial_image['index'])
            else:
                start = to_dt(df.index[0].to_pydatetime())
            image_start = initial_image.get('index', start)
            rtn[IMAGE_DOC] = {IMAGE_TIME: image_start, IMAGE: initial_image}
            final_image = TickStore._pandas_compute_final_image(df, initial_image, end)
        else:
            start = to_dt(df.index[0].to_pydatetime())
            final_image = {}
        rtn[END] = end
        rtn[START] = start

        logger.warning("NB treating all values as 'exists' - no longer sparse")
        rowmask = Binary(lz4_compressHC(np.packbits(np.ones(len(df), dtype='uint8')).tobytes()))

        index_name = df.index.names[0] or "index"
        if PD_VER < '0.23.0':
            recs = df.to_records(convert_datetime64=False)
        else:
            recs = df.to_records()

        for col in df:
            array = TickStore._ensure_supported_dtypes(recs[col])
            col_data = {
                DATA: Binary(lz4_compressHC(array.tobytes())),
                ROWMASK: rowmask,
                DTYPE: TickStore._str_dtype(array.dtype),
            }
            rtn[COLUMNS][col] = col_data
        rtn[INDEX] = Binary(
            lz4_compressHC(np.concatenate(
                ([recs[index_name][0].astype('datetime64[ms]').view('uint64')],
                 np.diff(
                     recs[index_name].astype('datetime64[ms]').view('uint64')))).tobytes()))
        return rtn, final_image

    @staticmethod
    def _to_bucket(ticks, symbol, initial_image):
        rtn = {SYMBOL: symbol, VERSION: CHUNK_VERSION_NUMBER, COLUMNS: {}, COUNT: len(ticks)}
        data = {}
        rowmask = {}
        start = to_dt(ticks[0]['index'])
        end = to_dt(ticks[-1]['index'])
        final_image = copy.copy(initial_image) if initial_image else {}
        for i, t in enumerate(ticks):
            if initial_image:
                final_image.update(t)
            for k, v in t.items():
                try:
                    if k != 'index':
                        rowmask[k][i] = 1
                    else:
                        v = TickStore._to_ms(v)
                        if data[k][-1] > v:
                            raise UnorderedDataException("Timestamps out-of-order: %s > %s" % (
                                ms_to_datetime(data[k][-1]), t))
                    data[k].append(v)
                except KeyError:
                    if k != 'index':
                        rowmask[k] = np.zeros(len(ticks), dtype='uint8')
                        rowmask[k][i] = 1
                    data[k] = [v]

        rowmask = dict([(k, Binary(lz4_compressHC(np.packbits(v).tobytes())))
                        for k, v in rowmask.items()])
        for k, v in data.items():
            if k != 'index':
                v = np.array(v)
                v = TickStore._ensure_supported_dtypes(v)
                rtn[COLUMNS][k] = {DATA: Binary(lz4_compressHC(v.tobytes())),
                                   DTYPE: TickStore._str_dtype(v.dtype),
                                   ROWMASK: rowmask[k]}

        if initial_image:
            image_start = initial_image.get('index', start)
            if image_start > start:
                raise UnorderedDataException("Image timestamp is after first tick: %s > %s" % (
                    image_start, start))
            start = min(start, image_start)
            rtn[IMAGE_DOC] = {IMAGE_TIME: image_start, IMAGE: initial_image}
        rtn[END] = end
        rtn[START] = start
        rtn[INDEX] = Binary(lz4_compressHC(np.concatenate(([data['index'][0]], np.diff(data['index']))).tobytes()))
        return rtn, final_image

    def max_date(self, symbol):
        """
        Return the maximum datetime stored for a particular symbol

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        """
        res = self._collection.find_one({SYMBOL: symbol}, projection={ID: 0, END: 1},
                                        sort=[(START, pymongo.DESCENDING)])
        if res is None:
            raise NoDataFoundException("No Data found for {}".format(symbol))
        return utc_dt_to_local_dt(res[END])

    def min_date(self, symbol):
        """
        Return the minimum datetime stored for a particular symbol

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        """
        res = self._collection.find_one({SYMBOL: symbol}, projection={ID: 0, START: 1},
                                        sort=[(START, pymongo.ASCENDING)])
        if res is None:
            raise NoDataFoundException("No Data found for {}".format(symbol))
        return utc_dt_to_local_dt(res[START])



import sys
from distutils.version import LooseVersion
from os import path

import pandas as pd
import pytest
from bson.binary import Binary
from bson.objectid import ObjectId
from mock import create_autospec, sentinel, Mock, call
import pickle

from arctic._compression import compress, compressHC
from arctic.exceptions import UnsupportedPickleStoreVersion
from arctic.store._pickle_store import PickleStore
from arctic.store._version_store_utils import checksum

PANDAS_VERSION = LooseVersion(pd.__version__)


def test_write():
    self = create_autospec(PickleStore)
    version = {}
    PickleStore.write(self, sentinel.arctic_lib, version, sentinel.symbol, 'item', sentinel.previous_version)
    assert version['data'] == 'item'


def test_write_object():
    arctic_lib = Mock()
    self = create_autospec(PickleStore)
    version = {'_id': ObjectId()}
    PickleStore.write(self, arctic_lib, version, 'sentinel.symbol', sentinel.item, sentinel.previous_version)
    assert 'data' not in version

    assert version['blob'] == '__chunked__V2'
    coll = arctic_lib.get_top_level_collection.return_value

    # Python 3.8 onwards uses protocol 5 which cannot be unpickled in Python versions below that, so limiting
    # it to use a maximum of protocol 4 in Python which is understood by 3.4 onwards and is still fairly efficient.
    # pickle version 4 is introduced with  python 3.4 and default with 3.8 onward
    pickle_protocol = min(4, pickle.HIGHEST_PROTOCOL)
    assert coll.update_one.call_args_list == [call({'sha': checksum('sentinel.symbol', {'segment': 0, 'data': Binary(compress(pickle.dumps(sentinel.item, pickle_protocol)))}),
                                                    'symbol': 'sentinel.symbol'},
                                                   {'$set': {'segment': 0, 'data': Binary(compress(pickle.dumps(sentinel.item, pickle_protocol)), 0)},
                                                    '$addToSet': {'parent': version['_id']}}, upsert=True)]


def test_read():
    self = create_autospec(PickleStore)
    version = {'data': 'item'}
    assert PickleStore.read(self, sentinel.arctic_lib, version, sentinel.symbol) == 'item'


def test_read_object_backwards_compat():
    self = create_autospec(PickleStore)
    version = {'blob': Binary(compressHC(pickle.dumps(object)))}
    assert PickleStore.read(self, sentinel.arctic_lib, version, sentinel.symbol) == object


def test_read_object_2():
    self = create_autospec(PickleStore)
    version = {'_id': sentinel._id,
               'blob': '__chunked__'}
    coll = Mock()
    arctic_lib = Mock()
    coll.find.return_value = [{'data': Binary(compressHC(pickle.dumps(object))),
                               'symbol': 'sentinel.symbol',
                               'segment': 1}
                              ]
    arctic_lib.get_top_level_collection.return_value = coll

    assert PickleStore.read(self, arctic_lib, version, sentinel.symbol) == object
    assert coll.find.call_args_list == [call({'symbol': sentinel.symbol, 'parent': sentinel._id})]


def test_read_with_base_version_id():
    self = create_autospec(PickleStore)
    version = {'_id': sentinel._id,
               'base_version_id': sentinel.base_version_id,
               'blob': '__chunked__'}
    coll = Mock()
    arctic_lib = Mock()
    coll.find.return_value = [{'data': Binary(compressHC(pickle.dumps(object))),
                               'symbol': 'sentinel.symbol',
                               'segment': 1}
                              ]
    arctic_lib.get_top_level_collection.return_value = coll

    assert PickleStore.read(self, arctic_lib, version, sentinel.symbol) == object
    assert coll.find.call_args_list == [call({'symbol': sentinel.symbol, 'parent': sentinel.base_version_id})]


@pytest.mark.xfail(sys.version_info >= (3,),
                   reason="lz4 data written with python2 not compatible with python3")
def test_read_backward_compatibility():
    """Test backwards compatibility with a pickled file that's created with Python 2.7.3,
    Numpy 1.7.1_ahl2 and Pandas 0.14.1
    """
    fname = path.join(path.dirname(__file__), "data", "test-data.pkl")

    # For newer versions; verify that unpickling fails when using cPickle
    if PANDAS_VERSION >= LooseVersion("0.16.1"):
        if sys.version_info[0] >= 3:
            with pytest.raises(UnicodeDecodeError), open(fname) as fh:
                pickle.load(fh)
        else:
            with pytest.raises(TypeError), open(fname) as fh:
                pickle.load(fh)

    # Verify that PickleStore() uses a backwards compatible unpickler.
    store = PickleStore()

    with open(fname) as fh:
        # PickleStore compresses data with lz4
        version = {'blob': compressHC(fh.read())}
    df = store.read(sentinel.arctic_lib, version, sentinel.symbol)

    expected = pd.DataFrame(range(4), pd.date_range(start="20150101", periods=4))
    assert (df == expected).all().all()


def test_unpickle_highest_protocol():
    """Pandas version 0.14.1 fails to unpickle a pandas.Series() in compat mode if the
    container has been pickled with HIGHEST_PROTOCOL.
    """
    version = {
        'blob': compressHC(pickle.dumps(pd.Series(), protocol=pickle.HIGHEST_PROTOCOL)),
    }

    store = PickleStore()
    ps = store.read(sentinel.arctic_lib, version, sentinel.symbol)

    expected = pd.Series()
    assert (ps == expected).all()


def test_pickle_chunk_V1_read():
    data = {'foo': b'abcdefghijklmnopqrstuvwxyz'}
    version = {'_id': sentinel._id,
               'blob': '__chunked__'}
    coll = Mock()
    arctic_lib = Mock()
    datap = compressHC(pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL))
    data_1 = datap[0:5]
    data_2 = datap[5:]
    coll.find.return_value = [{'data': Binary(data_1),
                               'symbol': 'sentinel.symbol',
                               'segment': 0},
                              {'data': Binary(data_2),
                               'symbol': 'sentinel.symbol',
                               'segment': 1},
                              ]
    arctic_lib.get_top_level_collection.return_value = coll

    ps = PickleStore()
    assert(data == ps.read(arctic_lib, version, sentinel.symbol))


def test_pickle_store_future_version():
    data = {'foo': b'abcdefghijklmnopqrstuvwxyz'}
    version = {'_id': sentinel._id,
               'blob': '__chunked__VERSION_ONE_MILLION'}
    coll = Mock()
    arctic_lib = Mock()
    datap = compressHC(pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL))
    data_1 = datap[0:5]
    data_2 = datap[5:]
    coll.find.return_value = [{'data': Binary(data_1),
                               'symbol': 'sentinel.symbol',
                               'segment': 0},
                              {'data': Binary(data_2),
                               'symbol': 'sentinel.symbol',
                               'segment': 1},
                              ]
    arctic_lib.get_top_level_collection.return_value = coll

    ps = PickleStore()
    with pytest.raises(UnsupportedPickleStoreVersion) as e:
        ps.read(arctic_lib, version, sentinel.symbol)
    assert('unsupported version of pickle store' in str(e.value))
